# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module bundles commonly used utility methods or helper classes that are used in multiple places within
OctoPrint's source code.
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

import io
import os
import traceback
import past.builtins
import sys
import re
import tempfile
import logging
import shutil
import threading
from functools import wraps
import warnings
import contextlib
import collections
import frozendict
import copy

from typing import Union

try:
	import queue
except ImportError:
	import Queue as queue

# noinspection PyCompatibility
from past.builtins import basestring, unicode

logger = logging.getLogger(__name__)


def to_bytes(s_or_u, encoding="utf-8", errors="strict"):
	# type: (Union[unicode, bytes], str, str) -> bytes
	"""Make sure ``s_or_u`` is a bytestring."""
	if s_or_u is None:
		return s_or_u

	if not isinstance(s_or_u, basestring):
		s_or_u = str(s_or_u)

	if isinstance(s_or_u, unicode):
		return s_or_u.encode(encoding, errors=errors)
	else:
		return s_or_u


def to_unicode(s_or_u, encoding="utf-8", errors="strict"):
	# type: (Union[unicode, bytes], str, str) -> unicode
	"""Make sure ``s_or_u`` is a unicode string."""
	if s_or_u is None:
		return s_or_u

	if not isinstance(s_or_u, basestring):
		s_or_u = str(s_or_u)

	if isinstance(s_or_u, bytes):
		return s_or_u.decode(encoding, errors=errors)
	else:
		return s_or_u


def to_native_str(s_or_u):
	# type: (Union[unicode, bytes]) -> str
	"""Make sure ``s_or_u`` is a 'str'."""
	if sys.version_info[0] == 2:
		return to_bytes(s_or_u)
	else:
		return to_unicode(s_or_u)


def sortable_value(value, default_value=""):
	if value is None:
		return default_value
	return value
sv = sortable_value


def pp(value):
	"""
	>>> pp(dict()) # doctest: +ALLOW_UNICODE
	'dict()'
	>>> pp(dict(a=1, b=2, c=3)) # doctest: +ALLOW_UNICODE
	'dict(a=1, b=2, c=3)'
	>>> pp(set()) # doctest: +ALLOW_UNICODE
	'set()'
	>>> pp({"a", "b"}) # doctest: +ALLOW_UNICODE
	"{'a', 'b'}"
	>>> pp(["a", "b", "d", "c"]) # doctest: +ALLOW_UNICODE
	"['a', 'b', 'd', 'c']"
	>>> pp(("a", "b", "d", "c")) # doctest: +ALLOW_UNICODE
	"('a', 'b', 'd', 'c')"
	>>> pp("foo") # doctest: +ALLOW_UNICODE
	"'foo'"
	>>> pp([dict(a=1, b=2), {"a", "c", "b"}, (1, 2), None, 1, True, "foo"]) # doctest: +ALLOW_UNICODE
	"[dict(a=1, b=2), {'a', 'b', 'c'}, (1, 2), None, 1, True, 'foo']"
	"""

	if isinstance(value, dict):
		# sort by keys
		r = "dict("
		r += ", ".join(map(lambda i: i[0] + "=" + pp(i[1]), sorted(value.items())))
		r += ")"
		return r
	elif isinstance(value, set):
		if len(value):
			# filled set: sort
			r = "{"
			r += ", ".join(map(pp, sorted(value)))
			r += "}"
			return r
		else:
			# empty set
			return "set()"
	elif isinstance(value, list):
		return "[" + ", ".join(map(pp, value)) + "]"
	elif isinstance(value, tuple):
		return "(" + ", ".join(map(pp, value)) + ")"
	else:
		return repr(value)


def warning_decorator_factory(warning_type):
	def specific_warning(message, stacklevel=1, since=None, includedoc=None, extenddoc=False):
		def decorator(func):
			@wraps(func)
			def func_wrapper(*args, **kwargs):
				# we need to increment the stacklevel by one because otherwise we'll get the location of our
				# func_wrapper in the log, instead of our caller (which is the real caller of the wrapped function)
				warnings.warn(message, warning_type, stacklevel=stacklevel + 1)
				return func(*args, **kwargs)

			if includedoc is not None and since is not None:
				docstring = "\n.. deprecated:: {since}\n   {message}\n\n".format(since=since, message=includedoc)
				if extenddoc and hasattr(func_wrapper, "__doc__") and func_wrapper.__doc__ is not None:
					docstring = func_wrapper.__doc__ + "\n" + docstring
				func_wrapper.__doc__ = docstring

			return func_wrapper

		return decorator
	return specific_warning


def warning_factory(warning_type):
	def specific_warning(message, stacklevel=1, since=None, includedoc=None, extenddoc=False):
		def decorator(o):
			def wrapper(f):
				def new(*args, **kwargs):
					warnings.warn(message, warning_type, stacklevel=stacklevel + 1)
					return f(*args, **kwargs)
				return new

			output = o.__class__.__new__(o.__class__, o)

			unwrappable_names = ("__weakref__", "__class__", "__dict__", "__doc__", "__str__", "__unicode__", "__repr__", "__getattribute__", "__setattr__")
			for method_name in dir(o):
				if method_name in unwrappable_names: continue

				setattr(output, method_name, wrapper(getattr(o, method_name)))

			if includedoc is not None and since is not None:
				docstring = "\n.. deprecated:: {since}\n   {message}\n\n".format(since=since, message=includedoc)
				if extenddoc and hasattr(wrapper, "__doc__") and wrapper.__doc__ is not None:
					docstring = wrapper.__doc__ + "\n" + docstring
					wrapper.__doc__ = docstring

			return output
		return decorator
	return specific_warning


deprecated = warning_decorator_factory(DeprecationWarning)
"""
A decorator for deprecated methods. Logs a deprecation warning via Python's `:mod:`warnings` module including the
supplied ``message``. The call stack level used (for adding the source location of the offending call to the
warning) can be overridden using the optional ``stacklevel`` parameter. If both ``since`` and ``includedoc`` are
provided, a deprecation warning will also be added to the function's docstring by providing or extending its ``__doc__``
property.

Arguments:
    message (string): The message to include in the deprecation warning.
    stacklevel (int): Stack level for including the caller of the offending method in the logged warning. Defaults to 1,
        meaning the direct caller of the method. It might make sense to increase this in case of the function call
        happening dynamically from a fixed position to not shadow the real caller (e.g. in case of overridden
        ``getattr`` methods).
    includedoc (string): Message about the deprecation to include in the wrapped function's docstring.
    extenddoc (boolean): If True the original docstring of the wrapped function will be extended by the deprecation
        message, if False (default) it will be replaced with the deprecation message.
    since (string): Version since when the function was deprecated, must be present for the docstring to get extended.

Returns:
    function: The wrapped function with the deprecation warnings in place.
"""

variable_deprecated = warning_factory(DeprecationWarning)
"""
A function for deprecated variables. Logs a deprecation warning via Python's `:mod:`warnings` module including the
supplied ``message``. The call stack level used (for adding the source location of the offending call to the
warning) can be overridden using the optional ``stacklevel`` parameter.

Arguments:
    message (string): The message to include in the deprecation warning.
    stacklevel (int): Stack level for including the caller of the offending method in the logged warning. Defaults to 1,
        meaning the direct caller of the method. It might make sense to increase this in case of the function call
        happening dynamically from a fixed position to not shadow the real caller (e.g. in case of overridden
        ``getattr`` methods).
    since (string): Version since when the function was deprecated, must be present for the docstring to get extended.

Returns:
    value: The value of the variable with the deprecation warnings in place.
"""

pending_deprecation = warning_decorator_factory(PendingDeprecationWarning)
"""
A decorator for methods pending deprecation. Logs a pending deprecation warning via Python's `:mod:`warnings` module
including the supplied ``message``. The call stack level used (for adding the source location of the offending call to
the warning) can be overridden using the optional ``stacklevel`` parameter. If both ``since`` and ``includedoc`` are
provided, a deprecation warning will also be added to the function's docstring by providing or extending its ``__doc__``
property.

Arguments:
    message (string): The message to include in the deprecation warning.
    stacklevel (int): Stack level for including the caller of the offending method in the logged warning. Defaults to 1,
        meaning the direct caller of the method. It might make sense to increase this in case of the function call
        happening dynamically from a fixed position to not shadow the real caller (e.g. in case of overridden
        ``getattr`` methods).
    extenddoc (boolean): If True the original docstring of the wrapped function will be extended by the deprecation
        message, if False (default) it will be replaced with the deprecation message.
    includedoc (string): Message about the deprecation to include in the wrapped function's docstring.
    since (string): Version since when the function was deprecated, must be present for the docstring to get extended.

Returns:
    function: The wrapped function with the deprecation warnings in place.
"""

variable_pending_deprecation = warning_factory(PendingDeprecationWarning)
"""
A decorator for variables pending deprecation. Logs a pending deprecation warning via Python's `:mod:`warnings` module
including the supplied ``message``. The call stack level used (for adding the source location of the offending call to
the warning) can be overridden using the optional ``stacklevel`` parameter.

Arguments:
    message (string): The message to include in the deprecation warning.
    stacklevel (int): Stack level for including the caller of the offending method in the logged warning. Defaults to 1,
        meaning the direct caller of the method. It might make sense to increase this in case of the function call
        happening dynamically from a fixed position to not shadow the real caller (e.g. in case of overridden
        ``getattr`` methods).
    since (string): Version since when the function was deprecated, must be present for the docstring to get extended.

Returns:
    value: The value of the variable with the deprecation warnings in place.
"""


to_str = deprecated("to_str has been renamed to to_bytes", since="1.3.11")(to_bytes)


def get_formatted_size(num):
	"""
	Formats the given byte count as a human readable rounded size expressed in the most pressing unit among B(ytes),
	K(ilo)B(ytes), M(ega)B(ytes), G(iga)B(ytes) and T(era)B(ytes), with one decimal place.

	Based on http://stackoverflow.com/a/1094933/2028598

	Arguments:
	    num (int): The byte count to format

	Returns:
	    string: The formatted byte count.
	"""

	for x in ["B","KB","MB","GB"]:
		if num < 1024.0:
			return "%3.1f%s" % (num, x)
		num /= 1024.0
	return "%3.1f%s" % (num, "TB")


def is_allowed_file(filename, extensions):
	"""
	Determines if the provided ``filename`` has one of the supplied ``extensions``. The check is done case-insensitive.

	Arguments:
	    filename (string): The file name to check against the extensions.
	    extensions (list): The extensions to check against, a list of strings

	Return:
	    boolean: True if the file name's extension matches one of the allowed extensions, False otherwise.
	"""

	return "." in filename and filename.rsplit(".", 1)[1].lower() in (x.lower() for x in extensions)


def get_formatted_timedelta(d):
	"""
	Formats a timedelta instance as "HH:MM:ss" and returns the resulting string.

	Arguments:
	    d (datetime.timedelta): The timedelta instance to format

	Returns:
	    string: The timedelta formatted as "HH:MM:ss"
	"""

	if d is None:
		return None
	hours = d.days * 24 + d.seconds // 3600
	minutes = (d.seconds % 3600) // 60
	seconds = d.seconds % 60
	return "%02d:%02d:%02d" % (hours, minutes, seconds)


def get_formatted_datetime(d):
	"""
	Formats a datetime instance as "YYYY-mm-dd HH:MM" and returns the resulting string.

	Arguments:
	    d (datetime.datetime): The datetime instance to format

	Returns:
	    string: The datetime formatted as "YYYY-mm-dd HH:MM"
	"""

	if d is None:
		return None

	return d.strftime("%Y-%m-%d %H:%M")


def get_class(name):
	"""
	Retrieves the class object for a given fully qualified class name.

	Arguments:
	    name (string): The fully qualified class name, including all modules separated by ``.``

	Returns:
	    type: The class if it could be found.

	Raises:
	    ImportError
	"""

	import importlib

	mod_name, cls_name = name.rsplit(".", 1)
	m = importlib.import_module(mod_name)
	try:
		return getattr(m, cls_name)
	except AttributeError:
		raise ImportError("No module named " + name)


def get_fully_qualified_classname(o):
	"""
	Returns the fully qualified class name for an object.

	Based on https://stackoverflow.com/a/2020083

	Args:
		o: the object of which to determine the fqcn

	Returns:
		(str) the fqcn of the object
	"""

	module = getattr(o.__class__, "__module__", None)
	if module is None:
		return o.__class__.__name__
	return module + "." + o.__class__.__name__


def get_exception_string():
	"""
	Retrieves the exception info of the last raised exception and returns it as a string formatted as
	``<exception type>: <exception message> @ <source file>:<function name>:<line number>``.

	Returns:
	    string: The formatted exception information.
	"""

	locationInfo = traceback.extract_tb(sys.exc_info()[2])[0]
	return "%s: '%s' @ %s:%s:%d" % (str(sys.exc_info()[0].__name__), str(sys.exc_info()[1]), os.path.basename(locationInfo[0]), locationInfo[2], locationInfo[1])


@deprecated("get_free_bytes has been deprecated and will be removed in the future",
            includedoc="Replaced by `psutil.disk_usage <http://pythonhosted.org/psutil/#psutil.disk_usage>`_.",
            since="1.2.5")
def get_free_bytes(path):
	import psutil
	return psutil.disk_usage(path).free


def get_dos_filename(input, existing_filenames=None, extension=None, whitelisted_extensions=None, **kwargs):
	"""
	Converts the provided input filename to a 8.3 DOS compatible filename. If ``existing_filenames`` is provided, the
	conversion result will be guaranteed not to collide with any of the filenames provided thus.

	Uses :func:`find_collision_free_name` internally.

	Arguments:
	    input (string): The original filename incl. extension to convert to the 8.3 format.
	    existing_filenames (list): A list of existing filenames with which the generated 8.3 name must not collide.
	        Optional.
	    extension (string): The .3 file extension to use for the generated filename. If not provided, the extension of
	        the provided ``filename`` will simply be truncated to 3 characters.
	    whitelisted_extensions (list): A list of extensions on ``input`` that will be left as-is instead of
	        exchanging for ``extension``.
	    kwargs (dict): Additional keyword arguments to provide to :func:`find_collision_free_name`.

	Returns:
	    string: A 8.3 compatible translation of the original filename, not colliding with the optionally provided
	        ``existing_filenames`` and with the provided ``extension`` or the original extension shortened to
	        a maximum of 3 characters.

	Raises:
	    ValueError: No 8.3 compatible name could be found that doesn't collide with the provided ``existing_filenames``.

	Examples:

	    >>> get_dos_filename("test1234.gco") # doctest: +ALLOW_UNICODE
	    'test1234.gco'
	    >>> get_dos_filename("test1234.gcode") # doctest: +ALLOW_UNICODE
	    'test1234.gco'
	    >>> get_dos_filename("test12345.gco") # doctest: +ALLOW_UNICODE
	    'test12~1.gco'
	    >>> get_dos_filename("test1234.fnord", extension="gco") # doctest: +ALLOW_UNICODE
	    'test1234.gco'
	    >>> get_dos_filename("auto0.g", extension="gco") # doctest: +ALLOW_UNICODE
	    'auto0.gco'
	    >>> get_dos_filename("auto0.g", extension="gco", whitelisted_extensions=["g"]) # doctest: +ALLOW_UNICODE
	    'auto0.g'
	    >>> get_dos_filename(None)
	    >>> get_dos_filename("foo") # doctest: +ALLOW_UNICODE
	    'foo'
	"""

	if input is None:
		return None

	if existing_filenames is None:
		existing_filenames = []

	if extension is not None:
		extension = extension.lower()

	if whitelisted_extensions is None:
		whitelisted_extensions = []

	filename, ext = os.path.splitext(input)

	ext = ext.lower()
	ext = ext[1:] if ext.startswith(".") else ext
	if ext in whitelisted_extensions or extension is None:
		extension = ext

	return find_collision_free_name(filename, extension, existing_filenames, **kwargs)


def find_collision_free_name(filename, extension, existing_filenames, max_power=2):
	"""
	Tries to find a collision free translation of "<filename>.<extension>" to the 8.3 DOS compatible format,
	preventing collisions with any of the ``existing_filenames``.

	First strips all of ``."/\\[]:;=,`` from the filename and extensions, converts them to lower case and truncates
	the ``extension`` to a maximum length of 3 characters.

	If the filename is already equal or less than 8 characters in length after that procedure and "<filename>.<extension>"
	are not contained in the ``existing_files``, that concatenation will be returned as the result.

	If not, the following algorithm will be applied to try to find a collision free name::

	    set counter := power := 1
	    while counter < 10^max_power:
	        set truncated := substr(filename, 0, 6 - power + 1) + "~" + counter
	        set result := "<truncated>.<extension>"
	        if result is collision free:
	            return result
	        counter++
	        if counter >= 10 ** power:
	            power++
	    raise ValueError

	This will basically -- for a given original filename of ``some_filename`` and an extension of ``gco`` -- iterate
	through names of the format ``some_f~1.gco``, ``some_f~2.gco``, ..., ``some_~10.gco``, ``some_~11.gco``, ...,
	``<prefix>~<n>.gco`` for ``n`` less than 10 ^ ``max_power``, returning as soon as one is found that is not colliding.

	Arguments:
	    filename (string): The filename without the extension to convert to 8.3.
	    extension (string): The extension to convert to 8.3 -- will be truncated to 3 characters if it's longer than
	        that.
	    existing_filenames (list): A list of existing filenames to prevent name collisions with.
	    max_power (int): Limits the possible attempts of generating a collision free name to 10 ^ ``max_power``
	        variations. Defaults to 2, so the name generation will maximally reach ``<name>~99.<ext>`` before
	        aborting and raising an exception.

	Returns:
	    string: A 8.3 representation of the provided original filename, ensured to not collide with the provided
	        ``existing_filenames``

	Raises:
	    ValueError: No collision free name could be found.

	Examples:

	    >>> find_collision_free_name("test1234", "gco", []) # doctest: +ALLOW_UNICODE
	    'test1234.gco'
	    >>> find_collision_free_name("test1234", "gcode", []) # doctest: +ALLOW_UNICODE
	    'test1234.gco'
	    >>> find_collision_free_name("test12345", "gco", []) # doctest: +ALLOW_UNICODE
	    'test12~1.gco'
	    >>> find_collision_free_name("test 123", "gco", []) # doctest: +ALLOW_UNICODE
	    'test_123.gco'
	    >>> find_collision_free_name("test1234", "g o", []) # doctest: +ALLOW_UNICODE
	    'test1234.g_o'
	    >>> find_collision_free_name("test12345", "gco", ["/test12~1.gco"]) # doctest: +ALLOW_UNICODE
	    'test12~2.gco'
	    >>> many_files = ["/test12~{}.gco".format(x) for x in range(10)[1:]]
	    >>> find_collision_free_name("test12345", "gco", many_files) # doctest: +ALLOW_UNICODE
	    'test1~10.gco'
	    >>> many_more_files = many_files + ["/test1~{}.gco".format(x) for x in range(10, 99)]
	    >>> find_collision_free_name("test12345", "gco", many_more_files) # doctest: +ALLOW_UNICODE
	    'test1~99.gco'
	    >>> many_more_files_plus_one = many_more_files + ["/test1~99.gco"]
	    >>> find_collision_free_name("test12345", "gco", many_more_files_plus_one) # doctest: +ALLOW_UNICODE
	    Traceback (most recent call last):
	    ...
	    ValueError: Can't create a collision free filename
	    >>> find_collision_free_name("test12345", "gco", many_more_files_plus_one, max_power=3) # doctest: +ALLOW_UNICODE
	    'test~100.gco'

	"""

	filename = to_unicode(filename)
	extension = to_unicode(extension)

	if filename.startswith("/"):
		filename = filename[1:]
	existing_filenames = [to_unicode(x[1:] if x.startswith("/") else x) for x in existing_filenames]

	def make_valid(text):
		return re.sub(r"\s+", "_", text.translate({ord(i):None for i in r".\"/\[]:;=,"})).lower()

	filename = make_valid(filename)
	extension = make_valid(extension)
	extension = extension[:3] if len(extension) > 3 else extension

	full_name_format = "{filename}.{extension}" if extension else "{filename}"

	result = full_name_format.format(filename=filename,
	                                 extension=extension)
	if len(filename) <= 8 and not result in existing_filenames:
		# early exit
		return result

	counter = 1
	power = 1
	prefix_format = "{segment}~{counter}"
	while counter < (10 ** max_power):
		prefix = prefix_format.format(segment=filename[:(6 - power + 1)], counter=str(counter))
		result = full_name_format.format(filename=prefix,
		                                 extension=extension)
		if result not in existing_filenames:
			return result
		counter += 1
		if counter >= 10 ** power:
			power += 1

	raise ValueError("Can't create a collision free filename")


def silent_remove(file):
	"""
	Silently removes a file. Does not raise an error if the file doesn't exist.

	Arguments:
	    file (string): The path of the file to be removed
	"""

	try:
		os.remove(file)
	except OSError:
		pass


def sanitize_ascii(line):
	if not isinstance(line, basestring):
		raise ValueError("Expected either str or unicode but got {} instead".format(line.__class__.__name__ if line is not None else None))
	return to_unicode(line, encoding="ascii", errors="replace").rstrip()


def filter_non_ascii(line):
	"""
	Filter predicate to test if a line contains non ASCII characters.

	Arguments:
	    line (string): The line to test

	Returns:
	    boolean: True if the line contains non ASCII characters, False otherwise.
	"""

	try:
		to_bytes(to_unicode(line, encoding="ascii"), encoding="ascii")
		return False
	except ValueError:
		return True


def chunks(l, n):
	"""
	Yield successive n-sized chunks from l.

	Taken from http://stackoverflow.com/a/312464/2028598
	"""
	for i in range(0, len(l), n):
		yield l[i:i+n]


def is_running_from_source():
	root = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../.."))
	return os.path.isdir(os.path.join(root, "src")) and os.path.isfile(os.path.join(root, "setup.py"))


def dict_merge(a, b, leaf_merger=None):
	"""
	Recursively deep-merges two dictionaries.

	Taken from https://www.xormedia.com/recursively-merge-dictionaries-in-python/

	Example::

	    >>> a = dict(foo="foo", bar="bar", fnord=dict(a=1))
	    >>> b = dict(foo="other foo", fnord=dict(b=2, l=["some", "list"]))
	    >>> expected = dict(foo="other foo", bar="bar", fnord=dict(a=1, b=2, l=["some", "list"]))
	    >>> dict_merge(a, b) == expected
	    True
	    >>> dict_merge(a, None) == a
	    True
	    >>> dict_merge(None, b) == b
	    True
	    >>> dict_merge(None, None) == dict()
	    True
	    >>> def leaf_merger(a, b):
	    ...     if isinstance(a, list) and isinstance(b, list):
	    ...         return a + b
	    ...     raise ValueError()
	    >>> result = dict_merge(dict(l1=[3, 4], l2=[1], a="a"), dict(l1=[1, 2], l2="foo", b="b"), leaf_merger=leaf_merger)
	    >>> result.get("l1") == [3, 4, 1, 2]
	    True
	    >>> result.get("l2") == "foo"
	    True
	    >>> result.get("a") == "a"
	    True
	    >>> result.get("b") == "b"
	    True

	Arguments:
	    a (dict): The dictionary to merge ``b`` into
	    b (dict): The dictionary to merge into ``a``
	    leaf_merger (callable): An optional callable to use to merge leaves (non-dict values)

	Returns:
	    dict: ``b`` deep-merged into ``a``
	"""

	from copy import deepcopy

	if a is None:
		a = dict()
	if b is None:
		b = dict()

	if not isinstance(b, dict):
		return b
	result = deepcopy(a)
	for k, v in b.items():
		if k in result and isinstance(result[k], dict):
			result[k] = dict_merge(result[k], v, leaf_merger=leaf_merger)
		else:
			merged = None
			if k in result and callable(leaf_merger):
				try:
					merged = leaf_merger(result[k], v)
				except ValueError:
					# can't be merged by leaf merger
					pass

			if merged is None:
				merged = deepcopy(v)

			result[k] = merged
	return result


def dict_sanitize(a, b):
	"""
	Recursively deep-sanitizes ``a`` based on ``b``, removing all keys (and
	associated values) from ``a`` that do not appear in ``b``.

	Example::

	    >>> a = dict(foo="foo", bar="bar", fnord=dict(a=1, b=2, l=["some", "list"]))
	    >>> b = dict(foo=None, fnord=dict(a=None, b=None))
	    >>> expected = dict(foo="foo", fnord=dict(a=1, b=2))
	    >>> dict_sanitize(a, b) == expected
	    True
	    >>> dict_clean(a, b) == expected
	    True

	Arguments:
	    a (dict): The dictionary to clean against ``b``.
	    b (dict): The dictionary containing the key structure to clean from ``a``.

	Results:
	    dict: A new dict based on ``a`` with all keys (and corresponding values) found in ``b`` removed.
	"""

	from copy import deepcopy
	if not isinstance(b, dict):
		return a

	result = deepcopy(a)
	for k, v in a.items():
		if not k in b:
			del result[k]
		elif isinstance(v, dict):
			result[k] = dict_sanitize(v, b[k])
		else:
			result[k] = deepcopy(v)
	return result
dict_clean = deprecated("dict_clean has been renamed to dict_sanitize",
                        includedoc="Replaced by :func:`dict_sanitize`")(dict_sanitize)


def dict_minimal_mergediff(source, target):
	"""
	Recursively calculates the minimal dict that would be needed to be deep merged with
	a in order to produce the same result as deep merging a and b.

	Example::

	    >>> a = dict(foo=dict(a=1, b=2), bar=dict(c=3, d=4))
	    >>> b = dict(bar=dict(c=3, d=5), fnord=None)
	    >>> c = dict_minimal_mergediff(a, b)
	    >>> c == dict(bar=dict(d=5), fnord=None)
	    True
	    >>> dict_merge(a, c) == dict_merge(a, b)
	    True

	Arguments:
	    source (dict): Source dictionary
	    target (dict): Dictionary to compare to source dictionary and derive diff for

	Returns:
	    dict: The minimal dictionary to deep merge on ``source`` to get the same result
	        as deep merging ``target`` on ``source``.
	"""

	if not isinstance(source, dict) or not isinstance(target, dict):
		raise ValueError("source and target must be dictionaries")

	if source == target:
		# shortcut: if both are equal, we return an empty dict as result
		return dict()

	from copy import deepcopy

	all_keys = set(list(source.keys()) + list(target.keys()))
	result = dict()
	for k in all_keys:
		if k not in target:
			# key not contained in b => not contained in result
			continue

		if k in source:
			# key is present in both dicts, we have to take a look at the value
			value_source = source[k]
			value_target = target[k]

			if value_source != value_target:
				# we only need to look further if the values are not equal

				if isinstance(value_source, dict) and isinstance(value_target, dict):
					# both are dicts => deeper down it goes into the rabbit hole
					result[k] = dict_minimal_mergediff(value_source, value_target)
				else:
					# new b wins over old a
					result[k] = deepcopy(value_target)

		else:
			# key is new, add it
			result[k] = deepcopy(target[k])
	return result


def dict_contains_keys(keys, dictionary):
	"""
	Recursively deep-checks if ``dictionary`` contains all keys found in ``keys``.

	Example::

	    >>> positive = dict(foo="some_other_bar", fnord=dict(b=100))
	    >>> negative = dict(foo="some_other_bar", fnord=dict(b=100, d=20))
	    >>> dictionary = dict(foo="bar", fnord=dict(a=1, b=2, c=3))
	    >>> dict_contains_keys(positive, dictionary)
	    True
	    >>> dict_contains_keys(negative, dictionary)
	    False

	Arguments:
	    a (dict): The dictionary to check for the keys from ``b``.
	    b (dict): The dictionary whose keys to check ``a`` for.

	Returns:
	    boolean: True if all keys found in ``b`` are also present in ``a``, False otherwise.
	"""

	if not isinstance(keys, dict) or not isinstance(dictionary, dict):
		return False

	for k, v in keys.items():
		if not k in dictionary:
			return False
		elif isinstance(v, dict):
			if not dict_contains_keys(v, dictionary[k]):
				return False

	return True


class fallback_dict(dict):
	def __init__(self, custom, *fallbacks):
		self.custom = custom
		self.fallbacks = fallbacks

	def __getitem__(self, item):
		for dictionary in self._all():
			if item in dictionary:
				return dictionary[item]
		raise KeyError()

	def __setitem__(self, key, value):
		self.custom[key] = value

	def __delitem__(self, key):
		# TODO: mark as deleted and leave fallbacks alone?
		for dictionary in self._all():
			if key in dictionary:
				del dictionary[key]

	def __contains__(self, key):
		return any((key in dictionary)
				   for dictionary in self._all())

	def keys(self):
		result = set()
		for dictionary in self._all():
			for k in dictionary.keys():
				if k in result:
					continue
				result.add(k)
				yield k

	def values(self):
		result = set()
		for dictionary in self._all():
			for k, v in dictionary.items():
				if k in result:
					continue
				result.add(k)
				yield k

	def items(self):
		result = set()
		for dictionary in self._all():
			for k, v in dictionary.items():
				if k in result:
					continue
				result.add(k)
				yield k, v

	def _all(self):
		yield self.custom
		for d in self.fallbacks:
			yield d


def dict_filter(dictionary, filter_function):
	"""
	Filters a dictionary with the provided filter_function

	Example::

	    >>> data = dict(key1="value1", key2="value2", other_key="other_value", foo="bar", bar="foo")
	    >>> dict_filter(data, lambda k, v: k.startswith("key")) == dict(key1="value1", key2="value2")
	    True
	    >>> dict_filter(data, lambda k, v: v.startswith("value")) == dict(key1="value1", key2="value2")
	    True
	    >>> dict_filter(data, lambda k, v: k == "foo" or v == "foo") == dict(foo="bar", bar="foo")
	    True
	    >>> dict_filter(data, lambda k, v: False) == dict()
	    True
	    >>> dict_filter(data, lambda k, v: True) == data
	    True
	    >>> dict_filter(None, lambda k, v: True)
	    Traceback (most recent call last):
	        ...
	    AssertionError
	    >>> dict_filter(data, None)
	    Traceback (most recent call last):
	        ...
	    AssertionError

	Arguments:
	    dictionary (dict): The dictionary to filter
	    filter_function (callable): The filter function to apply, called with key and
	        value of an entry in the dictionary, must return ``True`` for values to
	        keep and ``False`` for values to strip

	Returns:
	    dict: A shallow copy of the provided dictionary, stripped of the key-value-pairs
	        for which the ``filter_function`` returned ``False``
	"""
	assert isinstance(dictionary, dict)
	assert callable(filter_function)
	return dict((k, v) for k, v in dictionary.items() if filter_function(k, v))


# Source: http://stackoverflow.com/a/6190500/562769
class DefaultOrderedDict(collections.OrderedDict):
	def __init__(self, default_factory=None, *a, **kw):

		if default_factory is not None and not callable(default_factory):
			raise TypeError('first argument must be callable')
		collections.OrderedDict.__init__(self, *a, **kw)
		self.default_factory = default_factory

	def __getitem__(self, key):
		try:
			return collections.OrderedDict.__getitem__(self, key)
		except KeyError:
			return self.__missing__(key)

	def __missing__(self, key):
		if self.default_factory is None:
			raise KeyError(key)
		self[key] = value = self.default_factory()
		return value

	def __reduce__(self):
		if self.default_factory is None:
			args = tuple()
		else:
			args = self.default_factory,
		return type(self), args, None, None, list(self.items())

	def copy(self):
		return self.__copy__()

	def __copy__(self):
		return type(self)(self.default_factory, self)

	def __deepcopy__(self, memo):
		import copy
		return type(self)(self.default_factory,
		                  copy.deepcopy(list(self.items())))

	# noinspection PyMethodOverriding
	def __repr__(self):
		return 'OrderedDefaultDict(%s, %s)' % (self.default_factory,
		                                       collections.OrderedDict.__repr__(self))


class Object(object):
	pass

def interface_addresses(family=None):
	"""
	Retrieves all of the host's network interface addresses.
	"""

	import netifaces
	if not family:
		family = netifaces.AF_INET

	for interface in netifaces.interfaces():
		try:
			ifaddresses = netifaces.ifaddresses(interface)
		except Exception:
			continue
		if family in ifaddresses:
			for ifaddress in ifaddresses[family]:
				if not ifaddress["addr"].startswith("169.254."):
					yield ifaddress["addr"]

def address_for_client(host, port, timeout=3.05):
	"""
	Determines the address of the network interface on this host needed to connect to the indicated client host and port.
	"""

	for address in interface_addresses():
		try:
			if server_reachable(host, port, timeout=timeout, proto="udp", source=address):
				return address
		except Exception:
			continue

def server_reachable(host, port, timeout=3.05, proto="tcp", source=None):
	"""
	Checks if a server is reachable

	Args:
		host (str): host to check against
		port (int): port to check against
		timeout (float): timeout for check
		proto (str): ``tcp`` or ``udp``
		source (str): optional, socket used for check will be bound against this address if provided

	Returns:
		boolean: True if a connection to the server could be opened, False otherwise
	"""

	import socket

	if proto not in ("tcp", "udp"):
		raise ValueError("proto must be either 'tcp' or 'udp'")

	try:
		sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM if proto == "udp" else socket.SOCK_STREAM)
		sock.settimeout(timeout)
		if source is not None:
			sock.bind((source, 0))
		sock.connect((host, port))
		return True
	except Exception:
		return False

def guess_mime_type(data):
	import filetype
	return filetype.guess_mime(data)

def parse_mime_type(mime):
	import cgi

	if not mime or not isinstance(mime, basestring):
		raise ValueError("mime must be a non empty str or unicode")

	mime, params = cgi.parse_header(mime)

	if mime == "*":
		mime = "*/*"

	parts = mime.split("/") if "/" in mime else None
	if not parts or len(parts) != 2:
		raise ValueError("mime must be a mime type of format type/subtype")

	mime_type, mime_subtype = parts
	return mime_type.strip(), mime_subtype.strip(), params

def mime_type_matches(mime, other):
	if not isinstance(mime, tuple):
		mime = parse_mime_type(mime)
	if not isinstance(other, tuple):
		other = parse_mime_type(other)

	mime_type, mime_subtype, _ = mime
	other_type, other_subtype, _ = other

	type_matches = (mime_type == other_type or mime_type == "*" or other_type == "*")
	subtype_matches = (mime_subtype == other_subtype or mime_subtype == "*" or other_subtype == "*")

	return type_matches and subtype_matches

@contextlib.contextmanager
def atomic_write(filename, mode="w+b", encoding="utf-8", prefix="tmp", suffix="", permissions=0o644, max_permissions=0o777):
	if os.path.exists(filename):
		permissions |= os.stat(filename).st_mode
	permissions &= max_permissions

	# NamedTemporaryFile doesn't yet have an encoding parameter in py2, so we go the long way
	fd, path = tempfile.mkstemp(suffix=suffix, prefix=prefix)
	os.close(fd)

	if "b" in mode:
		temp_config = io.open(path, mode=mode)
	else:
		temp_config = io.open(path, mode=mode, encoding=encoding)

	try:
		yield temp_config
	finally:
		temp_config.close()
	os.chmod(temp_config.name, permissions)
	shutil.move(temp_config.name, filename)


@contextlib.contextmanager
def tempdir(ignore_errors=False, onerror=None, **kwargs):
	import tempfile
	import shutil

	dirpath = tempfile.mkdtemp(**kwargs)
	try:
		yield dirpath
	finally:
		shutil.rmtree(dirpath, ignore_errors=ignore_errors, onerror=onerror)


@contextlib.contextmanager
def temppath(prefix=None, suffix=""):
	import tempfile

	temp = tempfile.NamedTemporaryFile(prefix=prefix if prefix is not None else tempfile.template,
	                                   suffix=suffix,
	                                   delete=False)
	try:
		temp.close()
		yield temp.name
	finally:
		os.remove(temp.name)


def bom_aware_open(filename, encoding="ascii", mode="r", **kwargs):
	import codecs

	assert "b" not in mode, "binary mode not support by bom_aware_open"

	codec = codecs.lookup(encoding)
	encoding = codec.name

	if kwargs is None:
		kwargs = dict()

	potential_bom_attribute = "BOM_" + codec.name.replace("utf-", "utf").upper()
	if "r" in mode and hasattr(codecs, potential_bom_attribute):
		# these encodings might have a BOM, so let's see if there is one
		bom = getattr(codecs, potential_bom_attribute)

		with io.open(filename, mode='rb') as f:
			header = f.read(4)

		if header.startswith(bom):
			encoding += "-sig"

	return io.open(filename, encoding=encoding, mode=mode, **kwargs)


def is_hidden_path(path):
	if path is None:
		# we define a None path as not hidden here
		return False

	path = to_unicode(path)

	filename = os.path.basename(path)
	if filename.startswith("."):
		# filenames starting with a . are hidden
		return True

	if sys.platform == "win32":
		# if we are running on windows we also try to read the hidden file
		# attribute via the windows api
		try:
			import ctypes
			attrs = ctypes.windll.kernel32.GetFileAttributesW(path)
			assert attrs != -1     # INVALID_FILE_ATTRIBUTES == -1
			return bool(attrs & 2) # FILE_ATTRIBUTE_HIDDEN == 2
		except (AttributeError, AssertionError):
			pass

	# if we reach that point, the path is not hidden
	return False


try:
	from glob import escape
	glob_escape = escape
except ImportError:
	# no glob.escape - we need to implement our own
	_glob_escape_check = re.compile("([*?[])")
	_glob_escape_check_bytes = re.compile(b"([*?[])")

	def glob_escape(pathname):
		"""
		Ported from Python 3.4

		See https://github.com/python/cpython/commit/fd32fffa5ada8b8be8a65bd51b001d989f99a3d3
		"""

		drive, pathname = os.path.splitdrive(pathname)
		if isinstance(pathname, bytes):
			pathname = _glob_escape_check_bytes.sub(br"[\1]", pathname)
		else:
			pathname = _glob_escape_check.sub(r"[\1]", pathname)
		return drive + pathname

try:
	# py3
	from time import monotonic as monotonic_time
except ImportError:
	try:
		# py2 w/ suitable source for monotonic time
		from monotonic import monotonic as monotonic_time
	except RuntimeError:
		# no source of monotonic time available, nothing left but using time.time *cringe*
		import time
		monotonic_time = time.time


def thaw_frozendict(obj):
	if not isinstance(obj, (dict, frozendict.frozendict)):
		raise ValueError("obj must be a dict or frozendict instance")

	# only true love can thaw a frozen dict
	letitgo = dict()
	for key, value in obj.items():
		if isinstance(value, (dict, frozendict.frozendict)):
			letitgo[key] = thaw_frozendict(value)
		else:
			letitgo[key] = copy.deepcopy(value)
	return letitgo


def utmify(link, source=None, medium=None, name=None, term=None, content=None):
	if source is None:
		return link

	from collections import OrderedDict
	try:
		import urlparse
		from urllib import urlencode
	except ImportError:
		# python 3
		import urllib.parse as urlparse
		from urllib.parse import urlencode

	# inspired by https://stackoverflow.com/a/2506477
	parts = list(urlparse.urlparse(link))

	# parts[4] is the url query
	query = OrderedDict(urlparse.parse_qs(parts[4]))

	query["utm_source"] = source
	if medium is not None:
		query["utm_medium"] = medium
	if name is not None:
		query["utm_name"] = name
	if term is not None:
		query["utm_term"] = term
	if content is not None:
		query["utm_content"] = content

	parts[4] = urlencode(query, doseq=True)

	return urlparse.urlunparse(parts)


class RepeatedTimer(threading.Thread):
	"""
	This class represents an action that should be run repeatedly in an interval. It is similar to python's
	own :class:`threading.Timer` class, but instead of only running once the ``function`` will be run again and again,
	sleeping the stated ``interval`` in between.

	RepeatedTimers are started, as with threads, by calling their ``start()`` method. The timer can be stopped (in
	between runs) by calling the :func:`cancel` method. The interval the time waited before execution of a loop may
	not be exactly the same as the interval specified by the user.

	For example:

	.. code-block:: python

	   def hello():
	       print("Hello World!")

	   t = RepeatedTimer(1.0, hello)
	   t.start() # prints "Hello World!" every second

	Another example with dynamic interval and loop condition:

	.. code-block:: python

	   count = 0
	   maximum = 5
	   factor = 1

	   def interval():
	       global count
	       global factor
	       return count * factor

	   def condition():
	       global count
	       global maximum
	       return count <= maximum

	   def hello():
	       print("Hello World!")

	       global count
	       count += 1

	   t = RepeatedTimer(interval, hello, run_first=True, condition=condition)
	   t.start() # prints "Hello World!" 5 times, printing the first one
	             # directly, then waiting 1, 2, 3, 4s in between (adaptive interval)

	Arguments:
	    interval (float or callable): The interval between each ``function`` call, in seconds. Can also be a callable
	        returning the interval to use, in case the interval is not static.
	    function (callable): The function to call.
	    args (list or tuple): The arguments for the ``function`` call. Defaults to an empty list.
	    kwargs (dict): The keyword arguments for the ``function`` call. Defaults to an empty dict.
	    run_first (boolean): If set to True, the function will be run for the first time *before* the first wait period.
	        If set to False (the default), the function will be run for the first time *after* the first wait period.
	    condition (callable): Condition that needs to be True for loop to continue. Defaults to ``lambda: True``.
	    on_condition_false (callable): Callback to call when the timer finishes due to condition becoming false. Will
	        be called before the ``on_finish`` callback.
	    on_cancelled (callable): Callback to call when the timer finishes due to being cancelled. Will be called
	        before the ``on_finish`` callback.
	    on_finish (callable): Callback to call when the timer finishes, either due to being cancelled or since
	        the condition became false.
	    daemon (bool): daemon flag to set on underlying thread.
	"""

	def __init__(self, interval, function, args=None, kwargs=None,
	             run_first=False, condition=None, on_condition_false=None,
	             on_cancelled=None, on_finish=None, daemon=True):
		threading.Thread.__init__(self)

		if args is None:
			args = []
		if kwargs is None:
			kwargs = dict()
		if condition is None:
			condition = lambda: True

		if not callable(interval):
			self.interval = lambda: interval
		else:
			self.interval = interval

		self.function = function
		self.finished = threading.Event()
		self.args = args
		self.kwargs = kwargs
		self.run_first = run_first
		self.condition = condition

		self.on_condition_false = on_condition_false
		self.on_cancelled = on_cancelled
		self.on_finish = on_finish

		self.daemon = daemon

	def cancel(self):
		self._finish(self.on_cancelled)

	def run(self):
		while self.condition():
			if self.run_first:
				# if we are to run the function BEFORE waiting for the first time
				self.function(*self.args, **self.kwargs)

				# make sure our condition is still met before running into the downtime
				if not self.condition():
					break

			# wait, but break if we are cancelled
			self.finished.wait(self.interval())
			if self.finished.is_set():
				return

			if not self.run_first:
				# if we are to run the function AFTER waiting for the first time
				self.function(*self.args, **self.kwargs)

		# we'll only get here if the condition was false
		self._finish(self.on_condition_false)

	def _finish(self, *callbacks):
		self.finished.set()

		for callback in callbacks:
			if not callable(callback):
				continue
			callback()

		if callable(self.on_finish):
			self.on_finish()

class ResettableTimer(threading.Thread):
	"""
	This class represents an action that should be run after a specified amount of time. It is similar to python's
	own :class:`threading.Timer` class, with the addition of being able to reset the counter to zero.

	ResettableTimers are started, as with threads, by calling their ``start()`` method. The timer can be stopped (in
	between runs) by calling the :func:`cancel` method. Resetting the counter can be done with the :func:`reset` method.

	For example:

	.. code-block:: python

	   def hello():
	       print("Ran hello() at {}").format(time.time())

	   t = ResettableTimers(60.0, hello)
	   t.start()
	   print("Started at {}").format(time.time())
	   time.sleep(30)
	   t.reset()
	   print("Reset at {}").format(time.time())

	Arguments:
	    interval (float or callable): The interval before calling ``function``, in seconds. Can also be a callable
	        returning the interval to use, in case the interval is not static.
	    function (callable): The function to call.
	    args (list or tuple): The arguments for the ``function`` call. Defaults to an empty list.
	    kwargs (dict): The keyword arguments for the ``function`` call. Defaults to an empty dict.
	    on_cancelled (callable): Callback to call when the timer finishes due to being cancelled.
	    on_reset (callable): Callback to call when the timer is reset.
	"""

	def __init__(self, interval, function, args=None, kwargs=None, on_reset=None, on_cancelled=None):
		threading.Thread.__init__(self)
		self._event = threading.Event()
		self._mutex = threading.Lock()
		self.is_reset = True

		if args is None:
			args = []
		if kwargs is None:
			kwargs = dict()

		self.interval = interval
		self.function = function
		self.args = args
		self.kwargs = kwargs
		self.on_cancelled = on_cancelled
		self.on_reset = on_reset


	def run(self):
		while self.is_reset:
			with self._mutex:
				self.is_reset = False
			self._event.wait(self.interval)

		if not self._event.isSet():
			self.function(*self.args, **self.kwargs)
		with self._mutex:
			self._event.set()

	def cancel(self):
		with self._mutex:
			self._event.set()

		if callable(self.on_cancelled):
			self.on_cancelled()

	def reset(self, interval=None):
		with self._mutex:
			if interval:
				self.interval = interval

			self.is_reset = True
			self._event.set()
			self._event.clear()

		if callable(self.on_reset):
			self.on_reset()


class CountedEvent(object):

	def __init__(self, value=0, minimum=0, maximum=None, **kwargs):
		self._counter = 0
		self._min = minimum
		self._max = kwargs.get("max", maximum)
		self._mutex = threading.RLock()
		self._event = threading.Event()

		self._internal_set(value)

	@property
	def is_set(self):
		return self._event.is_set

	@property
	def counter(self):
		with self._mutex:
			return self._counter

	def set(self):
		with self._mutex:
			self._internal_set(self._counter + 1)

	def clear(self, completely=False):
		with self._mutex:
			if completely:
				self._internal_set(0)
			else:
				self._internal_set(self._counter - 1)

	def reset(self):
		self.clear(completely=True)

	def wait(self, timeout=None):
		self._event.wait(timeout)

	def blocked(self):
		return self.counter <= 0

	def acquire(self, blocking=1):
		return self._mutex.acquire(blocking=blocking)

	def release(self):
		return self._mutex.release()

	def _internal_set(self, value):
		self._counter = value
		if self._counter <= 0:
			if self._min is not None and self._counter < self._min:
				self._counter = self._min
			self._event.clear()
		else:
			if self._max is not None and self._counter > self._max:
				self._counter = self._max
			self._event.set()


class InvariantContainer(object):
	def __init__(self, initial_data=None, guarantee_invariant=None):
		try:
			from collections import Iterable
		except ImportError:
			# Python >= 3.8
			from collections.abc import Iterable
		from threading import RLock

		if guarantee_invariant is None:
			guarantee_invariant = lambda data: data

		self._data = []
		self._mutex = RLock()
		self._invariant = guarantee_invariant

		if initial_data is not None and isinstance(initial_data, Iterable):
			for item in initial_data:
				self.append(item)

	def append(self, item):
		with self._mutex:
			self._data.append(item)
			self._data = self._invariant(self._data)

	def remove(self, item):
		with self._mutex:
			self._data.remove(item)
			self._data = self._invariant(self._data)

	def __len__(self):
		return len(self._data)

	def __iter__(self):
		return self._data.__iter__()


class PrependableQueue(queue.Queue):

	def __init__(self, maxsize=0):
		queue.Queue.__init__(self, maxsize=maxsize)

	def prepend(self, item, block=True, timeout=True):
		from time import time as _time

		self.not_full.acquire()
		try:
			if self.maxsize > 0:
				if not block:
					if self._qsize() == self.maxsize:
						raise queue.Full
				elif timeout is None:
					while self._qsize() >= self.maxsize:
						self.not_full.wait()
				elif timeout < 0:
					raise ValueError("'timeout' must be a non-negative number")
				else:
					endtime = _time() + timeout
					while self._qsize() == self.maxsize:
						remaining = endtime - _time()
						if remaining <= 0.0:
							raise queue.Full
						self.not_full.wait(remaining)
			self._prepend(item)
			self.unfinished_tasks += 1
			self.not_empty.notify()
		finally:
			self.not_full.release()

	def _prepend(self, item):
		self.queue.appendleft(item)


class TypedQueue(PrependableQueue):

	def __init__(self, maxsize=0):
		PrependableQueue.__init__(self, maxsize=maxsize)
		self._lookup = set()

	def put(self, item, item_type=None, *args, **kwargs):
		PrependableQueue.put(self, (item, item_type), *args, **kwargs)

	def get(self, *args, **kwargs):
		item, _ = PrependableQueue.get(self, *args, **kwargs)
		return item

	def prepend(self, item, item_type=None, *args, **kwargs):
		PrependableQueue.prepend(self, (item, item_type), *args, **kwargs)

	def _put(self, item):
		_, item_type = item
		if item_type is not None:
			if item_type in self._lookup:
				raise TypeAlreadyInQueue(item_type, "Type {} is already in queue".format(item_type))
			else:
				self._lookup.add(item_type)

		PrependableQueue._put(self, item)

	def _get(self):
		item = PrependableQueue._get(self)
		_, item_type = item

		if item_type is not None:
			self._lookup.discard(item_type)

		return item

	def _prepend(self, item):
		_, item_type = item
		if item_type is not None:
			if item_type in self._lookup:
				raise TypeAlreadyInQueue(item_type, "Type {} is already in queue".format(item_type))
			else:
				self._lookup.add(item_type)

		PrependableQueue._prepend(self, item)

class TypeAlreadyInQueue(Exception):
	def __init__(self, t, *args, **kwargs):
		Exception.__init__(self, *args, **kwargs)
		self.type = t


class ConnectivityChecker(object):
	"""
	Regularly checks for online connectivity.

	Tries to open a connection to the provided ``host`` and ``port`` every ``interval``
	seconds and sets the ``online`` status accordingly.
	"""

	def __init__(self, interval, host, port, enabled=True, on_change=None):
		self._interval = interval
		self._host = host
		self._port = port
		self._enabled = enabled
		self._on_change = on_change

		self._logger = logging.getLogger(__name__ + ".connectivity_checker")

		# we initialize the online flag to True if we are not enabled (we don't know any better
		# but these days it's probably a sane default)
		self._online = not self._enabled

		self._check_worker = None
		self._check_mutex = threading.RLock()

		self._run()

	@property
	def online(self):
		"""Current online status, True if online, False if offline."""
		with self._check_mutex:
			return self._online

	@property
	def host(self):
		"""DNS host to query."""
		with self._check_mutex:
			return self._host

	@host.setter
	def host(self, value):
		with self._check_mutex:
			self._host = value

	@property
	def port(self):
		"""DNS port to query."""
		with self._check_mutex:
			return self._port

	@port.setter
	def port(self, value):
		with self._check_mutex:
			self._port = value

	@property
	def interval(self):
		"""Interval between consecutive automatic checks."""
		return self._interval

	@interval.setter
	def interval(self, value):
		self._interval = value

	@property
	def enabled(self):
		"""Whether the check is enabled or not."""
		return self._enabled

	@enabled.setter
	def enabled(self, value):
		with self._check_mutex:
			old_enabled = self._enabled
			self._enabled = value

			if not self._enabled:
				if self._check_worker is not None:
					self._check_worker.cancel()

				old_value = self._online
				self._online = True

				if old_value != self._online:
					self._trigger_change(old_value, self._online)

			elif self._enabled and not old_enabled:
				self._run()

	def check_immediately(self):
		"""Check immediately and return result."""
		with self._check_mutex:
			self._perform_check()
			return self.online

	def _run(self):
		from octoprint.util import RepeatedTimer

		if not self._enabled:
			return

		if self._check_worker is not None:
			self._check_worker.cancel()

		self._check_worker = RepeatedTimer(self._interval, self._perform_check,
		                                   run_first=True)
		self._check_worker.start()

	def _perform_check(self):
		if not self._enabled:
			return

		with self._check_mutex:
			self._logger.debug("Checking against {}:{} if we are online...".format(self._host, self._port))

			old_value = self._online
			self._online = server_reachable(self._host, port=self._port)

			if old_value != self._online:
				self._trigger_change(old_value, self._online)

	def _trigger_change(self, old_value, new_value):
		self._logger.info("Connectivity changed from {} to {}".format("online" if old_value else "offline",
		                                                              "online" if new_value else "offline"))
		if callable(self._on_change):
			self._on_change(old_value, new_value)


class CaseInsensitiveSet(collections.Set):
	"""
	Basic case insensitive set

	Any str or unicode values will be stored and compared in lower case. Other value types are left as-is.
	"""

	def __init__(self, *args):
		self.data = set(x.lower() if isinstance(x, past.builtins.basestring) else x for x in args)

	def __contains__(self, item):
		if isinstance(item, past.builtins.basestring):
			return item.lower() in self.data
		else:
			return item in self.data

	def __iter__(self):
		return iter(self.data)

	def __len__(self):
		return len(self.data)


# originally from https://stackoverflow.com/a/5967539
def natural_key(text):
	return [ int(c) if c.isdigit() else c for c in re.split(r"(\d+)", text) ]


def count(gen):
	"""Used instead of len(generator), which doesn't work"""
	n = 0
	for _ in gen:
		n += 1
	return n


def timing(f):
	@wraps(f)
	def decorator(*args, **kwargs):
		start = monotonic_time()
		try:
			return f(*args, **kwargs)
		finally:
			end = monotonic_time()
			logging.getLogger("octoprint.util.timing").debug("func:{} took {:0.2f}s".format(f.__name__,
			                                                                                end - start))
	return decorator


def generate_api_key():
	# noinspection PyCompatibility
	from builtins import bytes
	import uuid

	return ''.join('%02X' % z for z in bytes(uuid.uuid4().bytes))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import os

try:
	from os import scandir, walk
except ImportError:
	from scandir import scandir, walk

from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.loaders import FileSystemLoader, PrefixLoader, ChoiceLoader, \
	BaseLoader, TemplateNotFound, split_template_path

from webassets import Bundle

class FilteredFileSystemLoader(FileSystemLoader):
	"""
	Jinja2 ``FileSystemLoader`` subclass that allows filtering templates.

	Only such templates will be accessible for whose paths the provided
	``path_filter`` filter function returns True.

	``path_filter`` will receive the actual path on disc and should behave just
	like callables provided to Python's internal ``filter`` function, returning
	``True`` if the path is cleared and ``False`` if it is supposed to be removed
	from results and hence ``filter(path_filter, iterable)`` should be
	equivalent to ``[item for item in iterable if path_filter(item)]``.

	If ``path_filter`` is not set or not a ``callable``, the loader will
	behave just like the regular Jinja2 ``FileSystemLoader``.
	"""
	def __init__(self, searchpath, path_filter=None, **kwargs):
		FileSystemLoader.__init__(self, searchpath, **kwargs)
		self.path_filter = path_filter

	def get_source(self, environment, template):
		if callable(self.path_filter):
			pieces = split_template_path(template)
			if not self._combined_filter(os.path.join(*pieces)):
				raise TemplateNotFound(template)

		return FileSystemLoader.get_source(self, environment, template)

	def list_templates(self):
		result = FileSystemLoader.list_templates(self)

		if callable(self.path_filter):
			result = sorted(filter(self._combined_filter, result))

		return result

	def _combined_filter(self, path):
		filter_results = map(lambda x: not os.path.exists(os.path.join(x, path)) or self.path_filter(os.path.join(x, path)),
		                     self.searchpath)
		return all(filter_results)


class SelectedFilesLoader(BaseLoader):
	def __init__(self, files, encoding="utf-8"):
		self.files = files
		self.encoding = encoding

	def get_source(self, environment, template):
		if not template in self.files:
			raise TemplateNotFound(template)

		from jinja2.loaders import open_if_exists

		path = self.files[template]
		f = open_if_exists(path)
		if f is None:
			raise TemplateNotFound(template)
		try:
			contents = f.read().decode(self.encoding)
		finally:
			f.close()

		mtime = os.path.getmtime(path)

		def uptodate():
			try:
				return os.path.getmtime(path) == mtime
			except OSError:
				return False
		return contents, path, uptodate

	def list_templates(self):
		return self.files.keys()


class SelectedFilesWithConversionLoader(SelectedFilesLoader):
	def __init__(self, files, encoding="utf-8", conversion=None):
		SelectedFilesLoader.__init__(self, files, encoding=encoding)
		self.conversion = conversion

	def get_source(self, environment, template):
		contents = SelectedFilesLoader.get_source(self, environment, template)
		if callable(self.conversion):
			contents = self.conversion(contents[0]), contents[1], contents[2]
		return contents


def get_all_template_paths(loader):
	def walk_folder(folder):
		files = []
		walk_dir = walk(folder, followlinks=True)
		for dirpath, dirnames, filenames in walk_dir:
			for filename in filenames:
				path = os.path.join(dirpath, filename)
				files.append(path)
		return files

	def collect_templates_for_loader(loader):
		if isinstance(loader, SelectedFilesLoader):
			import copy
			return copy.copy(list(loader.files.values()))

		elif isinstance(loader, FilteredFileSystemLoader):
			result = []
			for folder in loader.searchpath:
				result += walk_folder(folder)
			return list(filter(loader.path_filter, result))

		elif isinstance(loader, FileSystemLoader):
			result = []
			for folder in loader.searchpath:
				result += walk_folder(folder)
			return result

		elif isinstance(loader, PrefixLoader):
			result = []
			for subloader in loader.mapping.values():
				result += collect_templates_for_loader(subloader)
			return result

		elif isinstance(loader, ChoiceLoader):
			result = []
			for subloader in loader.loaders:
				result += collect_templates_for_loader(subloader)
			return result

		return []

	return collect_templates_for_loader(loader)


def get_all_asset_paths(env):
	result = []

	def get_paths(bundle):
		r = []
		for content in bundle.resolve_contents():
			try:
				if not content:
					continue
				if isinstance(content[1], Bundle):
					r += get_paths(content[1])
				else:
					path = content[1]
					if not os.path.isfile(path):
						continue
					r.append(path)
			except IndexError:
				# intentionally ignored
				pass
		return r

	for bundle in env:
		result += get_paths(bundle)
	return result


class ExceptionHandlerExtension(Extension):
	tags = {"try"}

	def __init__(self, environment):
		super(ExceptionHandlerExtension, self).__init__(environment)
		self._logger = logging.getLogger(__name__)

	def parse(self, parser):
		token = next(parser.stream)
		lineno = token.lineno
		filename = parser.name
		error = parser.parse_expression()

		args = [error, nodes.Const(filename), nodes.Const(lineno)]
		try:
			body = parser.parse_statements(["name:endtry"], drop_needle=True)
			node = nodes.CallBlock(self.call_method("_handle_body", args),
			                       [], [], body).set_lineno(lineno)
		except Exception as e:
			# that was expected
			self._logger.exception("Caught exception while parsing template")
			node = nodes.CallBlock(self.call_method("_handle_error", [nodes.Const(self._format_error(error, e, filename, lineno))]),
			                       [], [], []).set_lineno(lineno)

		return node

	def _handle_body(self, error, filename, lineno, caller):
		try:
			return caller()
		except Exception as e:
			self._logger.exception("Caught exception while compiling template {filename} at line {lineno}".format(**locals()))
			error_string = self._format_error(error, e, filename, lineno)
			return error_string if error_string else ""

	def _handle_error(self, error, caller):
		return error if error else ""

	def _format_error(self, error, exception, filename, lineno):
		if not error:
			return ""

		try:
			return error.format(exception=exception, filename=filename, lineno=lineno)
		except Exception:
			self._logger.exception("Error while compiling exception output for template {filename} at line {lineno}".format(**locals()))
			return "Unknown error"

trycatch = ExceptionHandlerExtension


class MarkdownFilter(object):

	def __init__(self, app, **markdown_options):
		self._markdown_options = markdown_options
		app.jinja_env.filters.setdefault("markdown", self)

	def __call__(self, stream):
		from jinja2 import Markup
		from markdown import Markdown

		# Markdown is not thread safe
		markdown = Markdown(**self._markdown_options)
		return Markup(markdown.convert(stream))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


from octoprint.settings import settings


class PrintTimeEstimator(object):
	"""
	Estimator implementation.

	Subclass this and register via the octoprint.printer.estimation.factory hook to provide your own implementation.
	"""

	def __init__(self, job_type):
		self.stats_weighing_until = settings().getFloat(["estimation", "printTime", "statsWeighingUntil"])
		self.validity_range = settings().getFloat(["estimation", "printTime", "validityRange"])
		self.force_dumb_from_percent = settings().getFloat(["estimation", "printTime", "forceDumbFromPercent"])
		self.force_dumb_after_min = settings().getFloat(["estimation", "printTime", "forceDumbAfterMin"])

		threshold = None
		rolling_window = None
		countdown = None

		if job_type == "local" or job_type == "sdcard":
			# we are happy if the average of the estimates stays within 60s of the prior one
			threshold = settings().getFloat(["estimation", "printTime", "stableThreshold"])

			if job_type == "sdcard":
				# we are interesting in a rolling window of roughly the last 15s, so the number of entries has to be derived
				# by that divided by the sd status polling interval
				interval = settings().getFloat(["serial", "timeout", "sdStatus"])
				if interval <= 0:
					interval = 1.0
				rolling_window = int(15 // interval)
				if rolling_window < 1:
					rolling_window = 1

				# we are happy when one rolling window has been stable
				countdown = rolling_window

		self._data = TimeEstimationHelper(rolling_window=rolling_window, countdown=countdown, threshold=threshold)

	def estimate(self, progress, printTime, cleanedPrintTime, statisticalTotalPrintTime, statisticalTotalPrintTimeType):
		"""
		Tries to estimate the print time left for the print job

		This is somewhat horrible since accurate print time estimation is pretty much impossible to
		achieve, considering that we basically have only two data points (current progress in file and
		time needed for that so far - former prints or a file analysis might not have happened or simply
		be completely impossible e.g. if the file is stored on the printer's SD card) and
		hence can only do a linear estimation of a completely non-linear process. That's a recipe
		for inaccurate predictions right there. Yay.

		Anyhow, here's how this implementation works. This method gets the current progress in the
		printed file (percentage based on bytes read vs total bytes), the print time that elapsed,
		the same print time with the heat up times subtracted (if possible) and if available also
		some statistical total print time (former prints or a result from the GCODE analysis).

		  1. First get an "intelligent" estimate based on the :class:`~octoprint.printer.estimation.TimeEstimationHelper`.
		     That thing tries to detect if the estimation based on our progress and time needed for that becomes
		     stable over time through a rolling window and only returns a result once that appears to be the
		     case.
		  2. If we have any statistical data (former prints or a result from the GCODE analysis)
		     but no intelligent estimate yet, we'll use that for the next step. Otherwise, up to a certain percentage
		     in the print we do a percentage based weighing of the statistical data and the intelligent
		     estimate - the closer to the beginning of the print, the more precedence for the statistical
		     data, the closer to the cut off point, the more precedence for the intelligent estimate. This
		     is our preliminary total print time.
		  3. If the total print time is set, we do a sanity check for it. Based on the total print time
		     estimate and the time we already spent printing, we calculate at what percentage we SHOULD be
		     and compare that to the percentage at which we actually ARE. If it's too far off, our total
		     can't be trusted and we fall back on the dumb estimate. Same if the time we spent printing is
		     already higher than our total estimate.
		  4. If we do NOT have a total print time estimate yet but we've been printing for longer than
		     a configured amount of minutes or are further in the file than a configured percentage, we
		     also use the dumb estimate for now.

		Yes, all this still produces horribly inaccurate results. But we have to do this live during the print and
		hence can't produce to much computational overhead, we do not have any insight into the firmware implementation
		with regards to planner setup and acceleration settings, we might not even have access to the printed file's
		contents and such we need to find something that works "mostly" all of the time without costing too many
		resources. Feel free to propose a better solution within the above limitations (and I mean that, this solution
		here makes me unhappy).

		Args:
		    progress (float or None): Current percentage in the printed file
		    printTime (float or None): Print time elapsed so far
		    cleanedPrintTime (float or None): Print time elapsed minus the time needed for getting up to temperature
		        (if detectable).
		    statisticalTotalPrintTime (float or None): Total print time of past prints against same printer profile,
		        or estimated total print time from GCODE analysis.
		    statisticalTotalPrintTimeType (str or None): Type of statistical print time, either "average" (total time
		        of former prints) or "analysis"

		Returns:
		    (2-tuple) estimated print time left or None if not proper estimate could be made at all, origin of estimation
		"""

		if progress is None or progress == 0 or printTime is None or cleanedPrintTime is None:
			return None, None

		dumbTotalPrintTime = printTime / progress
		estimatedTotalPrintTime = self.estimate_total(progress, cleanedPrintTime)
		totalPrintTime = estimatedTotalPrintTime

		printTimeLeftOrigin = "estimate"
		if statisticalTotalPrintTime is not None:
			if estimatedTotalPrintTime is None:
				# no estimate yet, we'll use the statistical total
				totalPrintTime = statisticalTotalPrintTime
				printTimeLeftOrigin = statisticalTotalPrintTimeType

			else:
				if progress < self.stats_weighing_until:
					# still inside weighing range, use part stats, part current estimate
					sub_progress = progress * (1 / self.stats_weighing_until)
					if sub_progress > 1.0:
						sub_progress = 1.0
					printTimeLeftOrigin = "mixed-" + statisticalTotalPrintTimeType
				else:
					# use only the current estimate
					sub_progress = 1.0
					printTimeLeftOrigin = "estimate"

				# combine
				totalPrintTime = - sub_progress * statisticalTotalPrintTime \
				                 + sub_progress * estimatedTotalPrintTime

		printTimeLeft = None
		if totalPrintTime is not None:
			# sanity check current total print time estimate
			assumed_progress = cleanedPrintTime / totalPrintTime
			min_progress = progress - self.validity_range
			max_progress = progress + self.validity_range

			if min_progress <= assumed_progress <= max_progress and totalPrintTime > cleanedPrintTime:
				# appears sane, we'll use it
				printTimeLeft = totalPrintTime - cleanedPrintTime

			else:
				# too far from the actual progress or negative,
				# we use the dumb print time instead
				printTimeLeft = dumbTotalPrintTime - cleanedPrintTime
				printTimeLeftOrigin = "linear"

		else:
			printTimeLeftOrigin = "linear"
			if progress > self.force_dumb_from_percent or \
					cleanedPrintTime >= self.force_dumb_after_min * 60:
				# more than x% or y min printed and still no real estimate, ok, we'll use the dumb variant :/
				printTimeLeft = dumbTotalPrintTime - cleanedPrintTime

		if printTimeLeft is not None and printTimeLeft < 0:
			# shouldn't actually happen, but let's make sure
			printTimeLeft = None

		return printTimeLeft, printTimeLeftOrigin

	def estimate_total(self, progress, printTime):
		if not progress or not printTime or not self._data:
			return None
		else:
			return self._data.update(printTime / progress)


class TimeEstimationHelper(object):

	STABLE_THRESHOLD = 0.1
	STABLE_COUNTDOWN = 250
	STABLE_ROLLING_WINDOW = 250

	def __init__(self, rolling_window=None, countdown=None, threshold=None):
		if rolling_window is None:
			rolling_window = self.__class__.STABLE_ROLLING_WINDOW
		if countdown is None:
			countdown = self.__class__.STABLE_COUNTDOWN
		if threshold is None:
			threshold = self.__class__.STABLE_THRESHOLD

		self._rolling_window = rolling_window
		self._countdown = countdown
		self._threshold = threshold

		import collections
		self._distances = collections.deque([], self._rolling_window)
		self._totals = collections.deque([], self._rolling_window)
		self._sum_total = 0
		self._count = 0
		self._stable_counter = -1

	def is_stable(self):
		return self._stable_counter >= self._countdown

	def update(self, new_estimate):
		old_average_total = self.average_total

		self._sum_total += new_estimate
		self._totals.append(new_estimate)
		self._count += 1

		if old_average_total:
			self._distances.append(abs(self.average_total - old_average_total))

		if -self._threshold < self.average_distance < self._threshold:
			self._stable_counter += 1
		else:
			self._stable_counter = -1

		if self.is_stable():
			return self.average_total_rolling
		else:
			return None

	@property
	def average_total(self):
		if not self._count:
			return 0
		else:
			return self._sum_total / self._count

	@property
	def average_total_rolling(self):
		if not self._count or self._count < self._rolling_window or not len(self._totals):
			return -1
		else:
			return sum(self._totals) / len(self._totals)

	@property
	def average_distance(self):
		if not self._count or self._count < self._rolling_window + 1 or not len(self._distances):
			return -1
		else:
			return sum(self._distances) / len(self._distances)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <gina@octoprint.org>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import re

try:
	from urllib import parse as urlparse
except ImportError:
	import urlparse

from webassets.bundle import Bundle
from webassets.merge import MemoryHunk, BaseHunk
from webassets.filter import Filter
from webassets.filter.cssrewrite.base import PatternRewriter
import webassets.filter.cssrewrite.urlpath as urlpath


def replace_url(source_url, output_url, url):
	# If path is an absolute one, keep it
	parsed = urlparse.urlparse(url)

	if not parsed.scheme and not parsed.path.startswith('/'):
		abs_source_url = urlparse.urljoin(source_url, url)

		# relpath() will not detect this case
		if urlparse.urlparse(abs_source_url).scheme:
			return abs_source_url

		# rewritten url: relative path from new location (output)
		# to location of referenced file (source + current url)
		url = urlpath.relpath(output_url, abs_source_url)

	return url


class UrlRewriter(PatternRewriter):

	def input(self, _in, out, **kw):
		source, source_path, output, output_path = \
			kw['source'], kw['source_path'], kw['output'], kw['output_path']

		self.source_path = source_path
		self.output_path = output_path
		self.source_url = self.ctx.resolver.resolve_source_to_url(
			self.ctx, source_path, source)
		self.output_url = self.ctx.resolver.resolve_output_to_url(
			self.ctx, output)

		return super(UrlRewriter, self).input(_in, out, **kw)

	def replace_url(self, url):
		return replace_url(self.source_url, self.output_url, url)


class LessImportRewrite(UrlRewriter):
	name = "less_importrewrite"

	patterns = {
		"import_rewrite": re.compile(r"(@import(\s+\(.*\))?\s+)\"(.*)\";")
	}

	def import_rewrite(self, m):
		import_with_options = m.group(1)
		import_url = self.replace_url(m.group(3))
		return "{import_with_options}\"{import_url}\";".format(**locals())


class SourceMapRewrite(UrlRewriter):
	name = "sourcemap_urlrewrite"

	patterns = {
		"url_rewrite": re.compile(r"(//#\s+sourceMappingURL=)(.*)")
	}

	def url_rewrite(self, m):
		mapping = m.group(1)
		source_url = self.replace_url(m.group(2))
		return "{mapping}{source_url}".format(**locals())


class SourceMapRemove(PatternRewriter):
	name = "sourcemap_remove"

	patterns = {
		"sourcemap_remove": re.compile(r"(//#\s+sourceMappingURL=)(.*)")
	}

	def sourcemap_remove(self, m):
		return ""


class JsDelimiterBundler(Filter):
	name = "js_delimiter_bundler"
	options = {}

	def input(self, _in, out, **kwargs):
		source = kwargs.get("source", "n/a")

		out.write("// source: " + source + "\n")
		out.write(_in.read())
		out.write("\n;\n")


class ChainedHunk(BaseHunk):
	def __init__(self, *hunks):
		self._hunks = hunks

	def mtime(self):
		pass

	def data(self):
		result = ""
		for hunk in self._hunks:
			if isinstance(hunk, tuple) and len(hunk) == 2:
				hunk, f = hunk
			else:
				f = lambda x: x
			result += f(hunk.data())
		return result


_PLUGIN_BUNDLE_WRAPPER_PREFIX = \
"""// JS assets for plugin {plugin}
(function () {{
    try {{
        """
_PLUGIN_BUNDLE_WRAPPER_SUFFIX = \
"""
    }} catch (error) {{
        log.error("Error in JS assets for plugin {plugin}:", (error.stack || error));
    }}
}})();
"""
class JsPluginBundle(Bundle):
	def __init__(self, plugin, *args, **kwargs):
		Bundle.__init__(self, *args, **kwargs)
		self.plugin = plugin

	def _merge_and_apply(self, ctx, output, force, parent_debug=None,
	                     parent_filters=None, extra_filters=None,
	                     disable_cache=None):
		hunk = Bundle._merge_and_apply(self, ctx, output, force, parent_debug=parent_debug,
		                               parent_filters=parent_filters, extra_filters=extra_filters,
		                               disable_cache=disable_cache)

		return ChainedHunk(MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_PREFIX.format(plugin=self.plugin)),
		                   (hunk, lambda x: x.replace("\n", "\n        ")),
		                   MemoryHunk(_PLUGIN_BUNDLE_WRAPPER_SUFFIX.format(plugin=self.plugin)))



#!/usr/bin/env python2
# coding=utf-8

from setuptools import setup, find_packages
from distutils.command.build_py import build_py as _build_py
import os
import versioneer

import sys
sys.path.insert(0, os.path.join(os.path.dirname(os.path.realpath(__file__)), "src"))
import octoprint_setuptools
import setuptools

#-----------------------------------------------------------------------------------------------------------------------

# Supported python versions
# we test against 2.7, 3.6 and 3.7, so that's what we'll mark as supported
PYTHON_REQUIRES = ">=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4"

# Requirements for setup.py
SETUP_REQUIRES = [
	"markdown>=3.1,<3.2",
]

# Requirements for our application
INSTALL_REQUIRES = [
	# the following dependencies are non trivial to update since later versions introduce backwards incompatible
	# changes that might affect plugins, or due to other observed problems

	"flask>=0.12,<0.13",         # newer versions require newer Jinja versions
	"Jinja2>=2.8.1,<2.9",        # Jinja 2.9 has breaking changes WRT template scope - we can't
	                             # guarantee backwards compatibility for plugins and such with that
	                             # version, hence we need to pin to a lower version for now. See #1697
	"tornado==4.5.3",            # a memory leak was observed in tornado >= 5, see #2585
	"regex!=2018.11.6",          # avoid broken 2018.11.6. See #2874

	# anything below this should be checked on releases for new versions

	"Flask-Login>=0.4.1,<0.5",
	"Flask-Babel>=0.12,<0.13",
	"Flask-Assets>=0.12,<0.13",
	"blinker>=1.4,<1.5",         # dependency of the now vendor bundled flask_principal
	"werkzeug>=0.16,<0.17",
	"cachelib>=0.1,<0.2",
	"PyYAML>=5.1,<6",
	"markdown>=3.1,<3.2",
	"pyserial>=3.4,<3.5",
	"netaddr>=0.7.19,<0.8",
	"watchdog>=0.9.0,<0.10",
	"sarge==0.1.5post0",
	"netifaces>=0.10.9,<0.11",
	"pylru>=1.2,<1.3",
	"rsa>=4.0,<5",
	"pkginfo>=1.5.0.1,<1.6",
	"requests>=2.22.0,<3",
	"semantic_version>=2.8,<2.9",
	"psutil>=5.6.5,<5.7",
	"Click>=7,<8",
	"awesome-slugify>=1.6.5,<1.7",
	"feedparser>=5.2.1,<5.3",
	"future>=0.18.2,<0.19",
	"websocket-client>=0.56,<0.57",
	"wrapt>=1.11.2,<1.12",
	"emoji>=0.5.4,<0.6",
	"frozendict>=1.2,<1.3",
	"sentry-sdk==0.13.2",
	"filetype>=1.0.5,<2"
]

# Python 2 specific requirements
INSTALL_REQUIRES_PYTHON2 = [
	"futures>=3.3,<3.4",
	"monotonic>=1.5,<1.6",
	"scandir>=1.10,<1.11",
	"chainmap>=1.0.3,<1.1",
	"typing>=3.7.4.1,<4"
]

# OSX specific requirements
INSTALL_REQUIRES_OSX = [
	"appdirs>=1.4.3",
]

# Additional requirements for optional install options
EXTRA_REQUIRES = dict(
	# Dependencies for developing OctoPrint
	develop=[
		# Testing dependencies - SEE ALSO tox.ini
		"mock>=3.0.5,<4",
		"pytest==4.6.6",
		"pytest-doctest-custom>=1.0.0,<1.1",
		"ddt",

		# linter
		"flake8",

		# Documentation dependencies
		"sphinx>=1.8.5,<2",
		"sphinxcontrib-httpdomain",
		"sphinxcontrib-mermaid>=0.3.1",
		"sphinx_rtd_theme",
		"readthedocs-sphinx-ext==0.5.17" # Later versions require Jinja >= 2.9
	],

	# Dependencies for developing OctoPrint plugins
	plugins=[
		"cookiecutter>=1.6,<1.7"
	]
)

# Dependency links for any of the aforementioned dependencies
DEPENDENCY_LINKS = []

# adapted from https://hynek.me/articles/conditional-python-dependencies/
if int(setuptools.__version__.split(".", 1)[0]) < 18:
	# no bdist_wheel support for setuptools < 18 since we build universal wheels and our optional dependencies
	# would get lost there
	assert "bdist_wheel" not in sys.argv

	# add optional dependencies for setuptools versions < 18 that don't yet support environment markers
	if sys.version_info[0] < 3:
		INSTALL_REQUIRES += INSTALL_REQUIRES_PYTHON2

	if sys.platform == "darwin":
		INSTALL_REQUIRES += INSTALL_REQUIRES_OSX
else:
	# environment markers supported
	EXTRA_REQUIRES[":python_version < '3'"] = INSTALL_REQUIRES_PYTHON2
	EXTRA_REQUIRES[":sys_platform == 'darwin'"] = INSTALL_REQUIRES_OSX

#-----------------------------------------------------------------------------------------------------------------------
# Anything below here is just command setup and general setup configuration

here = os.path.abspath(os.path.dirname(__file__))

def read_file_contents(path):
	import codecs
	with codecs.open(path, encoding="utf-8") as f:
		return f.read()

def md_to_html_build_py_factory(files, baseclass):
	class md_to_html_build_py(baseclass):
		files = dict()

		def run(self):
			print("RUNNING md_to_html_build_py")
			if not self.dry_run:
				for directory, files in self.files.items():
					target_dir = os.path.join(self.build_lib, directory)
					self.mkpath(target_dir)

					for entry in files:
						if isinstance(entry, tuple):
							if len(entry) != 2:
								continue
							source, dest = entry[0], os.path.join(target_dir, entry[1])
						else:
							source = entry
							dest = os.path.join(target_dir, source + ".html")

						print("Rendering markdown from {} to {}".format(source, dest))

						from markdown import markdownFromFile
						markdownFromFile(input=source,
						                 output=dest,
						                 encoding="utf-8")
			baseclass.run(self)

	return type(md_to_html_build_py)(md_to_html_build_py.__name__,
	                                 (md_to_html_build_py,),
	                                 dict(files=files))

def get_cmdclass():
	cmdclass = versioneer.get_cmdclass()

	# add clean command
	cmdclass.update(dict(clean=octoprint_setuptools.CleanCommand.for_options(source_folder="src", eggs=["OctoPrint*.egg-info"])))

	# add translation commands
	translation_dir = "translations"
	pot_file = os.path.join(translation_dir, "messages.pot")
	bundled_dir = os.path.join("src", "octoprint", "translations")
	cmdclass.update(octoprint_setuptools.get_babel_commandclasses(pot_file=pot_file, output_dir=translation_dir, pack_name_prefix="OctoPrint-i18n-", pack_path_prefix="", bundled_dir=bundled_dir))

	cmdclass["build_py"] = md_to_html_build_py_factory({
		"octoprint/templates/_data": [
			"AUTHORS.md",
			"SUPPORTERS.md",
			"THIRDPARTYLICENSES.md",
		]
	}, cmdclass["build_py"] if "build_py" in cmdclass else _build_py)

	return cmdclass


def params():
	name = "OctoPrint"
	version = versioneer.get_version()
	cmdclass = get_cmdclass()

	description = "The snappy web interface for your 3D printer"
	long_description = read_file_contents(os.path.join(here, "README.md"))
	long_description_content_type = "text/markdown"

	python_requires = PYTHON_REQUIRES
	setup_requires = SETUP_REQUIRES
	install_requires = INSTALL_REQUIRES
	extras_require = EXTRA_REQUIRES
	dependency_links = DEPENDENCY_LINKS

	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"Environment :: Web Environment",
		"Framework :: Flask",
		"Intended Audience :: Developers",
		"Intended Audience :: Education",
		"Intended Audience :: End Users/Desktop",
		"Intended Audience :: Manufacturing",
		"Intended Audience :: Other Audience",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: GNU Affero General Public License v3",
		"Natural Language :: English",
		"Natural Language :: German",
		"Operating System :: OS Independent",
		"Programming Language :: Python",
		"Programming Language :: Python :: 2",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: Implementation :: CPython",
		"Programming Language :: JavaScript",
		"Topic :: Printing",
		"Topic :: System :: Monitoring"
	]
	author = "Gina Häußge"
	author_email = "gina@octoprint.org"
	url = "https://octoprint.org"
	license = "GNU Affero General Public License v3"
	keywords = "3dprinting 3dprinter 3d-printing 3d-printer octoprint"

	project_urls={
		"Community Forum": "https://community.octoprint.org",
		"Bug Reports": "https://github.com/foosel/OctoPrint/issues",
		"Source": "https://github.com/foosel/OctoPrint",
		"Funding": "https://donate.octoprint.org"
	}

	packages = find_packages(where="src")
	package_dir = {
		"": "src",
	}
	package_data = {
		"octoprint": octoprint_setuptools.package_data_dirs('src/octoprint',
		                                                    ['static', 'templates', 'plugins', 'translations'])
		             + ['util/piptestballoon/setup.py']
	}

	include_package_data = True
	zip_safe = False

	if os.environ.get('READTHEDOCS', None) == 'True':
		# we can't tell read the docs to please perform a pip install -e .[develop], so we help
		# it a bit here by explicitly adding the development dependencies, which include our
		# documentation dependencies
		install_requires = install_requires + extras_require['develop']

	entry_points = {
		"console_scripts": [
			"octoprint = octoprint:main"
		]
	}

	return locals()

setup(**params())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import collections
import logging
import sarge
import threading

from flask import request, make_response, jsonify, url_for
from flask_babel import gettext

import psutil

from octoprint.settings import settings as s

from octoprint.server import NO_CONTENT
from octoprint.server.api import api
from octoprint.server.util.flask import no_firstrun_access, get_remote_address
from octoprint.access.permissions import Permissions
from octoprint.logging import prefix_multilines

@api.route("/system/usage", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def readUsageForFolders():
	return jsonify(usage=_usageForFolders())

def _usageForFolders():
	data = {}
	for folder_name in s().get(['folder']).keys():
		path = s().getBaseFolder(folder_name, check_writable=False)
		if path is not None:
			usage = psutil.disk_usage(path)
			data[folder_name] = { 'free': usage.free, 'total': usage.total }
	return data

@api.route("/system", methods=["POST"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def performSystemAction():
	logging.getLogger(__name__).warn("Deprecated API call to /api/system made by {}, should be migrated to use /system/commands/custom/<action>".format(get_remote_address(request)))

	data = request.get_json(silent=True)
	if data is None:
		data = request.values

	if not "action" in data:
		return make_response(u"action to perform is not defined", 400)

	return executeSystemCommand("custom", data["action"])


@api.route("/system/commands", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def retrieveSystemCommands():
	return jsonify(core=_to_client_specs(_get_core_command_specs()),
	               custom=_to_client_specs(_get_custom_command_specs()))


@api.route("/system/commands/<string:source>", methods=["GET"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def retrieveSystemCommandsForSource(source):
	if source == "core":
		specs = _get_core_command_specs()
	elif source == "custom":
		specs = _get_custom_command_specs()
	else:
		return make_response(u"Unknown system command source: {}".format(source), 404)

	return jsonify(_to_client_specs(specs))


@api.route("/system/commands/<string:source>/<string:command>", methods=["POST"])
@no_firstrun_access
@Permissions.SYSTEM.require(403)
def executeSystemCommand(source, command):
	logger = logging.getLogger(__name__)

	if command == "divider":
		return make_response("Dividers cannot be executed", 400)

	command_spec = _get_command_spec(source, command)
	if not command_spec:
		return make_response(u"Command {}:{} not found".format(source, command), 404)

	if not "command" in command_spec:
		return make_response(u"Command {}:{} does not define a command to execute, can't proceed".format(source, command), 500)

	do_async = command_spec.get("async", False)
	do_ignore = command_spec.get("ignore", False)
	debug = command_spec.get("debug", False)

	if logger.isEnabledFor(logging.DEBUG) or debug:
		logger.info(u"Performing command for {}:{}: {}".format(source, command, command_spec["command"]))
	else:
		logger.info(u"Performing command for {}:{}".format(source, command))

	try:
		if "before" in command_spec and callable(command_spec["before"]):
			command_spec["before"]()
	except Exception as e:
		if not do_ignore:
			error = u"Command \"before\" for {}:{} failed: {}".format(source, command, e)
			logger.warning(error)
			return make_response(error, 500)

	try:
		def execute():
			# we run this with shell=True since we have to trust whatever
			# our admin configured as command and since we want to allow
			# shell-alike handling here...
			p = sarge.run(command_spec["command"],
			              stdout=sarge.Capture(),
			              stderr=sarge.Capture(),
			              shell=True)

			if not do_ignore and p.returncode != 0:
				returncode = p.returncode
				stdout_text = p.stdout.text
				stderr_text = p.stderr.text

				error = "Command for {}:{} failed with return code {}:\nSTDOUT: {}\nSTDERR: {}".format(source, command,
				                                                                                       returncode,
				                                                                                       stdout_text,
				                                                                                       stderr_text)
				logger.warning(prefix_multilines(error, prefix="! "))
				if not do_async:
					raise CommandFailed(error)

		if do_async:
			thread = threading.Thread(target=execute)
			thread.daemon = True
			thread.start()

		else:
			try:
				execute()
			except CommandFailed as exc:
				return make_response(exc.error, 500)

	except Exception as e:
		if not do_ignore:
			error = "Command for {}:{} failed: {}".format(source, command, e)
			logger.warning(error)
			return make_response(error, 500)

	return NO_CONTENT

def _to_client_specs(specs):
	result = list()
	for spec in specs.values():
		if not "action" in spec or not "source" in spec:
			continue
		copied = dict((k, v) for k, v in spec.items() if k in ("source", "action", "name", "confirm"))
		copied["resource"] = url_for(".executeSystemCommand",
		                             source=spec["source"],
		                             command=spec["action"],
		                             _external=True)
		result.append(copied)
	return result


def _get_command_spec(source, action):
	if source == "core":
		return _get_core_command_spec(action)
	elif source == "custom":
		return _get_custom_command_spec(action)
	else:
		return None


def _get_core_command_specs():
	def enable_safe_mode():
		s().set(["server", "startOnceInSafeMode"], True)
		s().save()

	commands = collections.OrderedDict(
		shutdown=dict(
			command=s().get(["server", "commands", "systemShutdownCommand"]),
			name=gettext("Shutdown system"),
			confirm=gettext(u"<strong>You are about to shutdown the system.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage).")),
		reboot=dict(
			command=s().get(["server", "commands", "systemRestartCommand"]),
			name=gettext("Reboot system"),
			confirm=gettext(u"<strong>You are about to reboot the system.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage).")),
		restart=dict(
			command=s().get(["server", "commands", "serverRestartCommand"]),
			name=gettext("Restart OctoPrint"),
			confirm=gettext(u"<strong>You are about to restart the OctoPrint server.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage).")),
		restart_safe=dict(
			command=s().get(["server", "commands", "serverRestartCommand"]),
			name=gettext("Restart OctoPrint in safe mode"),
			confirm=gettext(u"<strong>You are about to restart the OctoPrint server in safe mode.</strong></p><p>This action may disrupt any ongoing print jobs (depending on your printer's controller and general setup that might also apply to prints run directly from your printer's internal storage)."),
			before=enable_safe_mode)
	)

	available_commands = collections.OrderedDict()
	for action, spec in commands.items():
		if not spec["command"]:
			continue
		spec.update({'action': action, 'source': 'core', 'async': True, 'debug': True})
		available_commands[action] = spec
	return available_commands


def _get_core_command_spec(action):
	available_actions = _get_core_command_specs()
	if not action in available_actions:
		logging.getLogger(__name__).warn(u"Command for core action {} is not configured, you need to configure the command before it can be used".format(action))
		return None

	return available_actions[action]


def _get_custom_command_specs():
	specs = collections.OrderedDict()
	dividers = 0
	for spec in s().get(["system", "actions"]):
		if not "action" in spec:
			continue
		copied = dict(spec)
		copied["source"] = "custom"

		action = spec["action"]
		if action == "divider":
			dividers += 1
			action = "divider_{}".format(dividers)
		specs[action] = copied
	return specs


def _get_custom_command_spec(action):
	available_actions = _get_custom_command_specs()
	if not action in available_actions:
		return None

	return available_actions[action]


class CommandFailed(Exception):
	def __init__(self, error):
		self.error = error



# -*- coding: utf-8 -*-
"""
This module contains a functions that monkey patch third party dependencies in the one or other way.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

import functools
import warnings
import sys

def patch_sarge_async_on_py2():
	if sys.version_info[0] >= 3:
		return

	def move_async(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			if "async" in kwargs:
				warnings.warn("sarge's \"async\" parameter is deprecated. Use \"_async\" instead.",
				              DeprecationWarning,
				              stacklevel=2)
				kwargs["async_"] = kwargs.pop("async")
			return f(*args, **kwargs)
		return decorated_function

	import sarge

	# wrap helper
	sarge.run = move_async(sarge.run)

	# wrap Command
	sarge.Command.run = move_async(sarge.Command.run)

	# wrap Pipeline
	sarge.Pipeline.run = move_async(sarge.Pipeline.run)
	sarge.Pipeline.run_command_node = move_async(sarge.Pipeline.run_command_node)
	sarge.Pipeline.run_list_node = move_async(sarge.Pipeline.run_list_node)
	sarge.Pipeline.run_logical_node = move_async(sarge.Pipeline.run_logical_node)
	sarge.Pipeline.run_node = move_async(sarge.Pipeline.run_node)
	sarge.Pipeline.run_node_in_thread = move_async(sarge.Pipeline.run_node_in_thread)
	sarge.Pipeline.run_pipeline_node = move_async(sarge.Pipeline.run_pipeline_node)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"


import click
import logging

from octoprint import init_settings
from octoprint.cli import get_ctx_obj_option
from octoprint.access.groups import FilebasedGroupManager
from octoprint.access.users import FilebasedUserManager, UnknownUser, UserAlreadyExists
from octoprint.util import get_class, sv

click.disable_unicode_literals_warning = True

#~~ "octoprint user" commands


@click.group()
def user_commands():
	pass


@user_commands.group(name="user")
@click.pass_context
def user(ctx):
	"""
	User management.

	Note that this currently only supports managing user accounts stored in the configured user manager, not any
	user managers added through plugins and the "octoprint.users.factory" hook.
	"""
	try:
		logging.basicConfig(level=logging.DEBUG if get_ctx_obj_option(ctx, "verbosity", 0) > 0 else logging.WARN)
		settings = init_settings(get_ctx_obj_option(ctx, "basedir", None), get_ctx_obj_option(ctx, "configfile", None))

		group_manager_name = settings.get(["accessControl", "groupManager"])
		try:
			clazz = get_class(group_manager_name)
			group_manager = clazz()
		except AttributeError:
			click.echo("Could not instantiate group manager {}, "
			           "falling back to FilebasedGroupManager!".format(group_manager_name), err=True)
			group_manager = FilebasedGroupManager()

		ctx.obj.group_manager = group_manager

		name = settings.get(["accessControl", "userManager"])
		try:
			clazz = get_class(name)
			user_manager = clazz(group_manager=group_manager, settings=settings)
		except Exception:
			click.echo("Could not instantiate user manager {}, falling back to FilebasedUserManager!".format(name), err=True)
			user_manager = FilebasedUserManager(group_manager, settings=settings)
		finally:
			user_manager.enabled = settings.getBoolean(["accessControl", "enabled"])

		ctx.obj.user_manager = user_manager

	except Exception:
		click.echo("Could not instantiate user manager", err=True)
		return


@user.command(name="list")
@click.pass_context
def list_users_command(ctx):
	"""Lists user information"""
	users = ctx.obj.user_manager.get_all_users()
	_print_list(users)


@user.command(name="add")
@click.argument("username", type=click.STRING, required=True)
@click.password_option("--password", "password", help="Password for the user")
@click.option("-g", "--group", "groups", multiple=True,
              help="Groups to set on the user")
@click.option("-p", "--permission", "permissions", multiple=True,
              help="Individual permissions to set on the user")
@click.option("--admin", "is_admin", type=click.BOOL, is_flag=True, default=False,
              help="Adds user to admin group")
@click.pass_context
def add_user_command(ctx, username, password, groups, permissions, is_admin):
	"""Add a new user."""
	if not groups:
		groups = []

	if is_admin:
		groups.append(ctx.obj.group_manager.admin_group)

	try:
		ctx.obj.user_manager.add_user(username,
		                              password,
		                              groups=groups,
		                              permissions=permissions,
		                              active=True)

		user = ctx.obj.user_manager.find_user(username)
		if user:
			click.echo("User created:")
			click.echo("\t{}".format(_user_to_line(user.as_dict())))
	except UserAlreadyExists:
		click.echo("A user with the name {} does already exist!".format(username), err=True)


@user.command(name="remove")
@click.argument("username", type=click.STRING)
@click.pass_context
def remove_user_command(ctx, username):
	"""Remove an existing user."""
	confirm = click.prompt("This is will irreversibly destroy the user account! Enter 'yes' to confirm",
	                       type=click.STRING)

	if confirm.lower() == "yes":
		ctx.obj.user_manager.remove_user(username)
		click.echo("User {} removed.".format(username))
	else:
		click.echo("User {} not removed.".format(username))


@user.command(name="password")
@click.argument("username", type=click.STRING)
@click.password_option("--password", "password", help="New password for user")
@click.pass_context
def change_password_command(ctx, username, password):
	"""Change an existing user's password."""
	try:
		ctx.obj.user_manager.change_user_password(username, password)
		click.echo("Password changed for user {}.".format(username))
	except UnknownUser:
		click.echo("User {} does not exist!".format(username), err=True)


@user.command(name="activate")
@click.argument("username", type=click.STRING)
@click.pass_context
def activate_command(ctx, username):
	"""Activate a user account."""
	try:
		ctx.obj.user_manager.change_user_activation(username, True)
		click.echo("User {} activated.".format(username))

		user = ctx.obj.user_manager.find_user(username)
		if user:
			click.echo("User created:")
			click.echo("\t{}".format(_user_to_line(user.asDict())))
	except UnknownUser:
		click.echo("User {} does not exist!".format(username), err=True)


@user.command(name="deactivate")
@click.argument("username", type=click.STRING)
@click.pass_context
def deactivate_command(ctx, username):
	"""Activate a user account."""
	try:
		ctx.obj.user_manager.change_user_activation(username, False)
		click.echo("User {} activated.".format(username))

		user = ctx.obj.user_manager.find_user(username)
		if user:
			click.echo("User created:")
			click.echo("\t{}".format(_user_to_line(user.asDict())))
	except UnknownUser:
		click.echo("User {} does not exist!".format(username), err=True)


def _print_list(users):
	click.echo("{} users registered in the system:".format(len(users)))
	for user in sorted(map(lambda x: x.as_dict(), users), key=lambda x: sv(x.get("name"))):
		click.echo("\t{}".format(_user_to_line(user)))


def _user_to_line(user):
	return "{name}" \
	       "\n\t\tactive: {active}" \
	       "\n\t\tgroups: {groups}" \
	       "\n\t\tpermissions: {permissions}".format(name=user.get("name"),
	                                                 active=user.get("active", "False"),
	                                                 groups=", ".join(user.get("groups", [])),
	                                                 permissions=", ".join(user.get("permissions", [])))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.transports.pollingbase
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Polling transports base
"""

from octoprint.vendor.sockjs.tornado import basehandler
from octoprint.vendor.sockjs.tornado.transports import base


class PollingTransportBase(basehandler.PreflightHandler, base.BaseTransportMixin):
    """Polling transport handler base class"""
    def initialize(self, server):
        super(PollingTransportBase, self).initialize(server)

        self.session = None
        self.active = True

    def _get_session(self, session_id):
        return self.server.get_session(session_id)

    def _attach_session(self, session_id, start_heartbeat=False):
        session = self._get_session(session_id)

        if session is None:
            session = self.server.create_session(session_id)

        # Try to attach to the session
        if not session.set_handler(self, start_heartbeat):
            return False

        self.session = session

        # Verify if session is properly opened
        session.verify_state()

        return True

    def _detach(self):
        """Detach from the session"""
        if self.session:
            self.session.remove_handler(self)
            self.session = None

    def check_xsrf_cookie(self):
        pass

    def send_message(self, message, binary=False):
        """Called by the session when some data is available"""
        raise NotImplementedError()

    def session_closed(self):
        """Called by the session when it was closed"""
        self._detach()
        self.safe_finish()

    def on_connection_close(self):
        # If connection was dropped by the client, close session.
        # In all other cases, connection will be closed by the server.
        if self.session is not None:
            self.session.close(1002, 'Connection interrupted')

        super(PollingTransportBase, self).on_connection_close()

    def send_complete(self):
        self._detach()

        # Avoid race condition when waiting for write callback and session getting closed in between
        if not self._finished:
            self.safe_finish()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import pkg_resources

from octoprint.util.pip import PipCaller, UnknownPip
from octoprint.util.version import get_comparable_version
from .. import exceptions

logger = logging.getLogger("octoprint.plugins.softwareupdate.updaters.pip")
console_logger = logging.getLogger("octoprint.plugins.softwareupdate.updaters.pip.console")

_ALREADY_INSTALLED = "Requirement already satisfied (use --upgrade to upgrade)"
_POTENTIAL_EGG_PROBLEM_POSIX = "No such file or directory"
_POTENTIAL_EGG_PROBLEM_WINDOWS = "The system cannot find the file specified"

_pip_callers = dict()
_pip_version_dependency_links = pkg_resources.parse_version("1.5")

def can_perform_update(target, check, online=True):
	from .. import MINIMUM_PIP
	pip_caller = _get_pip_caller(command=check["pip_command"] if "pip_command" in check else None)
	return "pip" in check \
	       and pip_caller is not None \
	       and pip_caller.available \
	       and pip_caller.version >= get_comparable_version(MINIMUM_PIP) \
	       and (online or check.get("offline", False))

def _get_pip_caller(command=None):
	key = command
	if command is None:
		key = "__default"

	if not key in _pip_callers:
		try:
			_pip_callers[key] = PipCaller(configured=command)
		except UnknownPip:
			_pip_callers[key] = None

	return _pip_callers[key]

def perform_update(target, check, target_version, log_cb=None, online=True, force=False):
	pip_command = None
	if "pip_command" in check:
		pip_command = check["pip_command"]

	pip_working_directory = None
	if "pip_cwd" in check:
		pip_working_directory = check["pip_cwd"]

	if not online and not check.get("offline", False):
		raise exceptions.CannotUpdateOffline()

	pip_caller = _get_pip_caller(command=pip_command)
	if pip_caller is None:
		raise exceptions.UpdateError("Can't run pip", None)

	def _log_call(*lines):
		_log(lines, prefix=" ", stream="call")

	def _log_stdout(*lines):
		_log(lines, prefix=">", stream="stdout")

	def _log_stderr(*lines):
		_log(lines, prefix="!", stream="stderr")

	def _log_message(*lines):
		_log(lines, prefix="#", stream="message")

	def _log(lines, prefix=None, stream=None):
		if log_cb is None:
			return
		log_cb(lines, prefix=prefix, stream=stream)

	if log_cb is not None:
		pip_caller.on_log_call = _log_call
		pip_caller.on_log_stdout = _log_stdout
		pip_caller.on_log_stderr = _log_stderr

	install_arg = check["pip"].format(target_version=target_version, target=target_version)

	logger.debug("Target: %s, executing pip install %s" % (target, install_arg))
	pip_args = ["--disable-pip-version-check", "install", install_arg, "--no-cache-dir"]
	pip_kwargs = dict(env=dict(PYTHONWARNINGS=b"ignore:DEPRECATION::pip._internal.cli.base_command"))
	if pip_working_directory is not None:
		pip_kwargs.update(cwd=pip_working_directory)

	if "dependency_links" in check and check["dependency_links"]:
		pip_args += ["--process-dependency-links"]

	returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
	if returncode != 0:
		# This could be caused by an issue with pip/setuptools: If something (target or dependency of target) was
		# installed as an egg at an earlier date (e.g. thanks to just running python setup.py install), pip install
		# will throw an error after updating that something to a newer (non-egg) version since it will still have
		# the egg on its sys.path and expect to read data from it. Running the exact same command a second time
		# solved this during hunt for a workaround, so this is what we'll now do if it indeed looks like this
		# specific problem
		def is_egg_problem(line):
			return (_POTENTIAL_EGG_PROBLEM_POSIX in line or _POTENTIAL_EGG_PROBLEM_WINDOWS in line) and ".egg" in line

		if any(map(lambda x: is_egg_problem(x), stderr)) or any(map(lambda x: is_egg_problem(x), stdout)):
			_log_message("This looks like an error caused by a specific issue in upgrading Python \"eggs\"",
			             "via current versions of pip.",
			             "Performing a second install attempt as a work around.")
			returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
			if returncode != 0:
				raise exceptions.UpdateError("Error while executing pip install", (stdout, stderr))
		else:
			raise exceptions.UpdateError("Error while executing pip install", (stdout, stderr))

	if not force and any(map(lambda x: x.strip().startswith(_ALREADY_INSTALLED) and (install_arg in x or install_arg in x.lower()), stdout)):
		_log_message("Looks like we were already installed in this version. Forcing a reinstall.")
		force = True

	if force:
		logger.debug("Target: %s, executing pip install %s --ignore-reinstalled --force-reinstall --no-deps" % (target, install_arg))
		pip_args += ["--ignore-installed", "--force-reinstall", "--no-deps"]

		returncode, stdout, stderr = pip_caller.execute(*pip_args, **pip_kwargs)
		if returncode != 0:
			raise exceptions.UpdateError("Error while executing pip install --force-reinstall", (stdout, stderr))

	return "ok"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
    sockjs.tornado.session
    ~~~~~~~~~~~~~~~~~~~~~~

    SockJS session implementation.
"""

import logging

from octoprint.vendor.sockjs.tornado import sessioncontainer, periodic, proto
from octoprint.vendor.sockjs.tornado.util import bytes_to_str

LOG = logging.getLogger("tornado.general")


class ConnectionInfo(object):
    """Connection information object.

    Will be passed to the ``on_open`` handler of your connection class.

    Has few properties:

    `ip`
        Caller IP address
    `cookies`
        Collection of cookies
    `arguments`
        Collection of the query string arguments
    `headers`
        Collection of headers sent by the browser that established this
        connection
    `path`
        Request uri path
    """
    def __init__(self, ip, cookies, arguments, headers, path):
        self.ip = ip
        self.cookies = cookies
        self.arguments = arguments
        self.headers = headers
        self.path = path

    def get_argument(self, name):
        """Return single argument by name"""
        val = self.arguments.get(name)
        if val:
            return val[0]
        return None

    def get_cookie(self, name):
        """Return single cookie by its name"""
        return self.cookies.get(name)

    def get_header(self, name):
        """Return single header by its name"""
        return self.headers.get(name)


# Session states
CONNECTING = 0
OPEN = 1
CLOSING = 2
CLOSED = 3


class BaseSession(object):
    """Base session implementation class"""
    def __init__(self, conn, server):
        """Base constructor.

        `conn`
            Connection class
        `server`
            SockJSRouter instance
        """
        self.server = server
        self.stats = server.stats

        self.send_expects_json = False

        self.handler = None
        self.state = CONNECTING

        self.conn_info = None

        self.conn = conn(self)

        self.close_reason = None

    def set_handler(self, handler):
        """Set transport handler
        ``handler``
            Handler, should derive from the `sockjs.tornado.transports.base.BaseTransportMixin`.
        """
        if self.handler is not None:
            raise Exception('Attempted to overwrite BaseSession handler')

        self.handler = handler
        self.transport_name = self.handler.name

        if self.conn_info is None:
            self.conn_info = handler.get_conn_info()
            self.stats.on_sess_opened(self.transport_name)

        return True

    def verify_state(self):
        """Verify if session was not yet opened. If it is, open it and call connections `on_open`"""
        if self.state == CONNECTING:
            self.state = OPEN

            self.conn.on_open(self.conn_info)

    def remove_handler(self, handler):
        """Remove active handler from the session

        `handler`
            Handler to remove
        """
        # Attempt to remove another handler
        if self.handler != handler:
            raise Exception('Attempted to remove invalid handler')

        self.handler = None

    def close(self, code=3000, message='Go away!'):
        """Close session or endpoint connection.

        `code`
            Closing code
        `message`
            Close message
        """
        if self.state != CLOSED:
            try:
                self.conn.on_close()
            except Exception:
                LOG.debug("Failed to call on_close().", exc_info=True)
            finally:
                self.state = CLOSED
                self.close_reason = (code, message)
                self.conn = None

            # Bump stats
            self.stats.on_sess_closed(self.transport_name)

            # If we have active handler, notify that session was closed
            if self.handler is not None:
                self.handler.session_closed()

    def delayed_close(self):
        """Delayed close - won't close immediately, but on next ioloop tick."""
        self.state = CLOSING
        self.server.io_loop.add_callback(self.close)

    def get_close_reason(self):
        """Return last close reason tuple.

        For example:

            if self.session.is_closed:
                code, reason = self.session.get_close_reason()

        """
        if self.close_reason:
            return self.close_reason

        return (3000, 'Go away!')

    @property
    def is_closed(self):
        """Check if session was closed."""
        return self.state == CLOSED or self.state == CLOSING

    def send_message(self, msg, stats=True, binary=False):
        """Send or queue outgoing message

        `msg`
            Message to send
        `stats`
            If set to True, will update statistics after operation completes
        """
        raise NotImplementedError()

    def send_jsonified(self, msg, stats=True):
        """Send or queue outgoing message which was json-encoded before. Used by the `broadcast`
        method.

        `msg`
            JSON-encoded message to send
        `stats`
            If set to True, will update statistics after operation completes
        """
        raise NotImplementedError()

    def broadcast(self, clients, msg):
        """Optimized `broadcast` implementation. Depending on type of the session, will json-encode
        message once and will call either `send_message` or `send_jsonifed`.

        `clients`
            Clients iterable
        `msg`
            Message to send
        """
        self.server.broadcast(clients, msg)


class Session(BaseSession, sessioncontainer.SessionMixin):
    """SockJS session implementation.
    """

    def __init__(self, conn, server, session_id, expiry=None):
        """Session constructor.

        `conn`
            Default connection class
        `server`
            `SockJSRouter` instance
        `session_id`
            Session id
        `expiry`
            Session expiry time
        """
        # Initialize session
        sessioncontainer.SessionMixin.__init__(self, session_id, expiry)
        BaseSession.__init__(self, conn, server)

        self.send_queue = ''
        self.send_expects_json = True

        # Heartbeat related stuff
        self._heartbeat_timer = None
        self._heartbeat_interval = self.server.settings['heartbeat_delay'] * 1000

        self._immediate_flush = self.server.settings['immediate_flush']
        self._pending_flush = False

        self._verify_ip = self.server.settings['verify_ip']

    # Session callbacks
    def on_delete(self, forced):
        """Session expiration callback

        `forced`
            If session item explicitly deleted, forced will be set to True. If
            item expired, will be set to False.
        """
        # Do not remove connection if it was not forced and there's running connection
        if not forced and self.handler is not None and not self.is_closed:
            self.promote()
        else:
            self.close()

    # Add session
    def set_handler(self, handler, start_heartbeat=True):
        """Set active handler for the session

        `handler`
            Associate active Tornado handler with the session
        `start_heartbeat`
            Should session start heartbeat immediately
        """
        # Check if session already has associated handler
        if self.handler is not None:
            handler.send_pack(proto.disconnect(2010, "Another connection still open"))
            return False

        if self._verify_ip and self.conn_info is not None:
            # If IP address doesn't match - refuse connection
            if handler.request.remote_ip != self.conn_info.ip:
                LOG.error('Attempted to attach to session %s (%s) from different IP (%s)' % (
                              self.session_id,
                              self.conn_info.ip,
                              handler.request.remote_ip
                              ))

                handler.send_pack(proto.disconnect(2010, "Attempted to connect to session from different IP"))
                return False

        if (self.state == CLOSING or self.state == CLOSED) and not self.send_queue:
            handler.send_pack(proto.disconnect(*self.get_close_reason()))
            return False

        # Associate handler and promote session
        super(Session, self).set_handler(handler)

        self.promote()

        if start_heartbeat:
            self.start_heartbeat()

        return True

    def verify_state(self):
        """Verify if session was not yet opened. If it is, open it and call connections `on_open`"""
        # If we're in CONNECTING state - send 'o' message to the client
        if self.state == CONNECTING:
            self.handler.send_pack(proto.CONNECT)

        # Call parent implementation
        super(Session, self).verify_state()

    def remove_handler(self, handler):
        """Detach active handler from the session

        `handler`
            Handler to remove
        """
        super(Session, self).remove_handler(handler)

        self.promote()
        self.stop_heartbeat()

    def send_message(self, msg, stats=True, binary=False):
        """Send or queue outgoing message

        `msg`
            Message to send
        `stats`
            If set to True, will update statistics after operation completes
        """
        self.send_jsonified(proto.json_encode(bytes_to_str(msg)), stats)

    def send_jsonified(self, msg, stats=True):
        """Send JSON-encoded message

        `msg`
            JSON encoded string to send
        `stats`
            If set to True, will update statistics after operation completes
        """
        msg = bytes_to_str(msg)

        if self._immediate_flush:
            if self.handler and self.handler.active and not self.send_queue:
                # Send message right away
                self.handler.send_pack('a[%s]' % msg)
            else:
                if self.send_queue:
                    self.send_queue += ','
                self.send_queue += msg

                self.flush()
        else:
            if self.send_queue:
                self.send_queue += ','
            self.send_queue += msg

            if not self._pending_flush:
                self.server.io_loop.add_callback(self.flush)
                self._pending_flush = True

        if stats:
            self.stats.on_pack_sent(1)

    def flush(self):
        """Flush message queue if there's an active connection running"""
        self._pending_flush = False

        if self.handler is None or not self.handler.active or not self.send_queue:
            return

        self.handler.send_pack('a[%s]' % self.send_queue)
        self.send_queue = ''

    def close(self, code=3000, message='Go away!'):
        """Close session.

        `code`
            Closing code
        `message`
            Closing message
        """
        if self.state != CLOSED:
            # Notify handler
            if self.handler is not None:
                self.handler.send_pack(proto.disconnect(code, message))

        super(Session, self).close(code, message)

    # Heartbeats
    def start_heartbeat(self):
        """Reset hearbeat timer"""
        self.stop_heartbeat()

        self._heartbeat_timer = periodic.Callback(self._heartbeat,
                                                  self._heartbeat_interval,
                                                  self.server.io_loop)
        self._heartbeat_timer.start()

    def stop_heartbeat(self):
        """Stop active heartbeat"""
        if self._heartbeat_timer is not None:
            self._heartbeat_timer.stop()
            self._heartbeat_timer = None

    def delay_heartbeat(self):
        """Delay active heartbeat"""
        if self._heartbeat_timer is not None:
            self._heartbeat_timer.delay()

    def _heartbeat(self):
        """Heartbeat callback"""
        if self.handler is not None:
            self.handler.send_pack(proto.HEARTBEAT)
        else:
            self.stop_heartbeat()

    def on_messages(self, msg_list):
        """Handle incoming messages

        `msg_list`
            Message list to process
        """
        self.stats.on_pack_recv(len(msg_list))

        for msg in msg_list:
            if self.state == OPEN:
                self.conn.on_message(msg)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

from future.utils import with_metaclass
from flask import g, abort
from flask_babel import gettext
from octoprint.vendor.flask_principal import Permission, PermissionDenied, RoleNeed, Need

from functools import wraps
from collections import OrderedDict, defaultdict

# noinspection PyCompatibility
from past.builtins import basestring

from octoprint.access import ADMIN_GROUP, USER_GROUP, READONLY_GROUP


class OctoPrintPermission(Permission):
	@classmethod
	def convert_needs_to_dict(cls, needs):
		ret_needs = defaultdict(list)
		for need in needs:
			if need.value not in ret_needs[need.method]:
				ret_needs[need.method].append(need.value)
		return ret_needs

	@classmethod
	def convert_to_needs(cls, needs):
		result = []
		for need in needs:
			# noinspection PyCompatibility
			if isinstance(need, Need):
				result.append(need)
			elif isinstance(need, Permission):
				result += need.needs
			elif isinstance(need, basestring):
				result.append(RoleNeed(need))
		return result

	def __init__(self, name, description, *needs, **kwargs):
		self._name = name
		self._description = description
		self._dangerous = kwargs.pop("dangerous", False)
		self._default_groups = kwargs.pop("default_groups", [])

		self._key = None

		Permission.__init__(self, *self.convert_to_needs(needs))

	def as_dict(self):
		return dict(
			key=self.key,
			name=self.get_name(),
			dangerous=self._dangerous,
			default_groups=self._default_groups,
			description=self.get_description(),
			needs=self.convert_needs_to_dict(self.needs)
		)

	@property
	def key(self):
		return self._key

	@key.setter
	def key(self, value):
		self._key = value

	@property
	def default_groups(self):
		return self._default_groups

	def get_name(self):
		return self._name

	def get_description(self):
		return self._description

	def allows(self, identity):
		"""Whether the identity can access this permission.
		Overridden from Permission.allows to make sure the Identity provides ALL
		required needs instead of ANY required need.

		:param identity: The identity
		"""
		if self.needs and len(self.needs.intersection(identity.provides)) != len(self.needs):
			return False

		if self.excludes and self.excludes.intersection(identity.provides):
			return False

		return True


	def union(self, other):
		"""Create a new OctoPrintPermission with the requirements of the union of this
		and other.

		:param other: The other permission
		"""
		p = self.__class__(self._name, self._description, *self.needs.union(other.needs))
		p.excludes.update(self.excludes.union(other.excludes))
		return p

	def difference(self, other):
		"""Create a new OctoPrintPermission consisting of requirements in this
		permission and not in the other.
		"""

		p = self.__class__(self._name, self._description, *self.needs.difference(other.needs))
		p.excludes.update(self.excludes.difference(other.excludes))
		return p

	def __repr__(self):
		return '{}({!r}, {!r}, {})'.format(self.__class__.__name__,
		                                   self.get_name(),
		                                   self.get_description(),
		                                   ', '.join(map(repr, self.needs)))

	def __hash__(self):
		return self.get_name().__hash__()

	def __eq__(self, other):
		return isinstance(other, OctoPrintPermission) and other.get_name() == self.get_name()


class PluginOctoPrintPermission(OctoPrintPermission):

	def __init__(self, *args, **kwargs):
		self.plugin = kwargs.pop("plugin", None)
		OctoPrintPermission.__init__(self, *args, **kwargs)

	def as_dict(self):
		result = OctoPrintPermission.as_dict(self)
		result["plugin"] = self.plugin
		return result


class PluginIdentityContext(object):
	"""Identity context for not initialized Permissions

	Needed to support @Permissions.PLUGIN_X_Y.require()

	Will search the permission when needed
	"""
	def __init__(self, key, http_exception=None):
		self.key = key
		self.http_exception = http_exception
		"""The permission of this principal
		"""

	@property
	def identity(self):
		"""The identity of this principal
		"""
		return g.identity

	def can(self):
		"""Whether the identity has access to the permission
		"""
		permission = getattr(Permissions, self.key)
		if permission is None or isinstance(permission, PluginPermissionDecorator):
			raise UnknownPermission(self.key)

		return permission.can()

	def __call__(self, f):
		@wraps(f)
		def _decorated(*args, **kw):
			with self:
				rv = f(*args, **kw)
			return rv

		return _decorated

	def __enter__(self):
		permission = getattr(Permissions, self.key)
		if permission is None or isinstance(permission, PluginPermissionDecorator):
			raise UnknownPermission(self.key)

		# check the permission here
		if not permission.can():
			if self.http_exception:
				abort(self.http_exception, permission)
			raise PermissionDenied(permission)

	def __exit__(self, *args):
		return False


class PluginPermissionDecorator(Permission):
	"""Decorator Class for not initialized Permissions

	Needed to support @Permissions.PLUGIN_X_Y.require()
	"""
	def __init__(self, key):
		self.key = key

	def require(self, http_exception=None):
		return PluginIdentityContext(self.key, http_exception)


class PermissionsMetaClass(type):
	permissions = OrderedDict()

	def __new__(mcs, name, bases, args):
		cls = type.__new__(mcs, name, bases, args)

		for key, value in args.items():
			if isinstance(value, OctoPrintPermission):
				value.key = key
				mcs.permissions[key] = value
				delattr(cls, key)

		return cls

	def __setattr__(cls, key, value):
		if isinstance(value, OctoPrintPermission):
			if key in cls.permissions:
				raise PermissionAlreadyExists(key)
			value.key = key
			cls.permissions[key] = value

	def __getattr__(cls, key):
		permission = cls.permissions.get(key)

		if key.startswith("PLUGIN_") and permission is None:
			return PluginPermissionDecorator(key)

		return permission

	def all(cls):
		return list(cls.permissions.values())

	def filter(cls, cb):
		return list(filter(cb, cls.all()))

	def find(cls, p, filter=None):
		key = None
		if isinstance(p, OctoPrintPermission):
			key = p.key
		elif isinstance(p, dict):
			key = p.get("key")
		elif isinstance(p, basestring):
			key = p

		if key is None:
			return None

		return cls.match(lambda p: p.key == key, filter=filter)

	def match(cls, match, filter=None):
		if callable(filter):
			permissions = cls.filter(filter)
		else:
			permissions = cls.all()

		for permission in permissions:
			if match(permission):
				return permission

		return None


class Permissions(with_metaclass(PermissionsMetaClass)):

	# Special permission
	ADMIN                  = OctoPrintPermission("Admin",
	                                             gettext("Admin is allowed to do everything"),
	                                             RoleNeed("admin"),
	                                             dangerous=True,
	                                             default_groups=[ADMIN_GROUP])

	STATUS                 = OctoPrintPermission("Status",
	                                             gettext("Allows to gather basic status information, e.g. job progress, "
	                                                     "printer state, temperatures, ... Mandatory for the default UI "
	                                                     "to work"),
	                                             RoleNeed("status"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])

	CONNECTION             = OctoPrintPermission("Connection",
	                                             gettext("Allows to connect to and disconnect from a printer"),
	                                             RoleNeed("connection"),
	                                             default_groups=[USER_GROUP])

	WEBCAM                 = OctoPrintPermission("Webcam",
	                                             gettext("Allows to watch the webcam stream"),
	                                             RoleNeed("webcam"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])

	SYSTEM                 = OctoPrintPermission("System",
	                                             gettext("Allows to run system commands, e.g. restart OctoPrint, "
	                                                     "shutdown or reboot the system"),
	                                             RoleNeed("system"),
	                                             dangerous=True)

	FILES_LIST             = OctoPrintPermission("File List",
	                                             gettext("Allows to retrieve a list of all uploaded files, including"
	                                                     "their metadata (e.g. date, file size, analysis results, ...)"),
	                                             RoleNeed("files_list"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])
	FILES_UPLOAD           = OctoPrintPermission("File Upload",
	                                             gettext("Allows users to upload new files"),
	                                             RoleNeed("files_upload"),
	                                             default_groups=[USER_GROUP])
	FILES_DOWNLOAD         = OctoPrintPermission("File Download",
	                                             gettext("Allows users to download files. The GCODE viewer is "
	                                                     "affected by this as well."),
	                                             RoleNeed("files_download"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])
	FILES_DELETE           = OctoPrintPermission("File Delete",
	                                             gettext("Allows users to delete files"),
	                                             RoleNeed("files_delete"),
	                                             default_groups=[USER_GROUP])
	FILES_SELECT           = OctoPrintPermission("File Select",
	                                             gettext("Allows to select a file for printing"),
	                                             RoleNeed("files_select"),
	                                             default_groups=[USER_GROUP])

	PRINT                  = OctoPrintPermission("Print",
	                                             gettext("Allows to start, pause and cancel a print job"),
	                                             RoleNeed("print"),
	                                             default_groups=[USER_GROUP])

	GCODE_VIEWER           = OctoPrintPermission("GCODE viewer",
	                                             gettext("Allows access to the GCODE viewer if the \"File Download\""
	                                                     "permission is also set."),
	                                             RoleNeed("gcodeviewer"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])

	MONITOR_TERMINAL       = OctoPrintPermission("Terminal",
	                                             gettext("Allows to watch the terminal tab but not to send commands "
	                                                     "to the printer from it"),
	                                             RoleNeed("monitor_terminal"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])

	CONTROL                = OctoPrintPermission("Control",
	                                             gettext("Allows to control of the printer by using the temperature controls,"
	                                                     "the control tab or sending commands through the terminal."),
	                                             RoleNeed("control"),
	                                             default_groups=[USER_GROUP])

	SLICE                  = OctoPrintPermission("Slice",
	                                             gettext("Allows to slice files"),
	                                             RoleNeed("slice"),
	                                             default_groups=[USER_GROUP])

	TIMELAPSE_LIST         = OctoPrintPermission("Timelapse List",
	                                             gettext("Allows to list timelapse videos"),
	                                             RoleNeed("timelapse_list"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])
	TIMELAPSE_DOWNLOAD     = OctoPrintPermission("Timelapse Download",
	                                             gettext("Allows to download timelapse videos"),
	                                             RoleNeed("timelapse_download"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])
	TIMELAPSE_DELETE       = OctoPrintPermission("Timelapse Delete",
	                                             gettext("Allows to delete timelapse videos and unrendered timelapses"),
	                                             RoleNeed("timelapse_delete"),
	                                             default_groups=[USER_GROUP])
	TIMELAPSE_ADMIN        = OctoPrintPermission("Timelapse Admin",
	                                             gettext("Allows to change the timelapse settings and delete or "
	                                                     "render unrendered timelapses. Includes the \"Timelapse List\","
	                                                     "\"Timelapse Delete\" and \"Timelapse Download\" permissions"),
	                                             RoleNeed("timelapse_admin"), TIMELAPSE_LIST, TIMELAPSE_DOWNLOAD,
	                                             default_groups=[USER_GROUP])

	SETTINGS_READ          = OctoPrintPermission("Settings Access",
	                                             gettext("Allows to read non sensitive settings. Mandatory for the "
	                                                     "default UI to work."),
	                                             RoleNeed("settings_read"),
	                                             default_groups=[USER_GROUP, READONLY_GROUP])
	SETTINGS               = OctoPrintPermission("Settings Admin",
	                                             gettext("Allows to manage settings and also to read sensitive settings"),
	                                             RoleNeed("settings"),
	                                             dangerous=True)



class PermissionAlreadyExists(Exception):
	def __init__(self, permission):
		Exception.__init__(self, "Permission %s already exists" % permission)


class UnknownPermission(Exception):
	def __init__(self, permissionname):
		Exception.__init__(self, "Unknown permission: %s" % permissionname)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask import request, jsonify, make_response

from octoprint.settings import settings
from octoprint.server import printer, printerProfileManager, NO_CONTENT
from octoprint.server.api import api
from octoprint.server.util.flask import no_firstrun_access, get_json_command_from_request
from octoprint.access.permissions import Permissions

@api.route("/connection", methods=["GET"])
@Permissions.STATUS.require(403)
def connectionState():
	state, port, baudrate, printer_profile = printer.get_current_connection()
	current = {
		"state": state,
		"port": port,
		"baudrate": baudrate,
		"printerProfile": printer_profile["id"] if printer_profile is not None and "id" in printer_profile else "_default"
	}

	return jsonify({"current": current, "options": _get_options()})


@api.route("/connection", methods=["POST"])
@no_firstrun_access
@Permissions.CONNECTION.require(403)
def connectionCommand():
	valid_commands = {
		"connect": [],
		"disconnect": [],
		"fake_ack": []
	}

	command, data, response = get_json_command_from_request(request, valid_commands)
	if response is not None:
		return response

	if command == "connect":
		connection_options = printer.__class__.get_connection_options()

		port = None
		baudrate = None
		printerProfile = None
		if "port" in data:
			port = data["port"]
			if port not in connection_options["ports"] and port != "AUTO":
				return make_response("Invalid port: %s" % port, 400)
		if "baudrate" in data:
			baudrate = data["baudrate"]
			if baudrate not in connection_options["baudrates"] and baudrate != 0:
				return make_response("Invalid baudrate: %d" % baudrate, 400)
		if "printerProfile" in data:
			printerProfile = data["printerProfile"]
			if not printerProfileManager.exists(printerProfile):
				return make_response("Invalid printer profile: %s" % printerProfile, 400)
		if "save" in data and data["save"]:
			settings().set(["serial", "port"], port)
			settings().setInt(["serial", "baudrate"], baudrate)
			printerProfileManager.set_default(printerProfile)
		if "autoconnect" in data:
			settings().setBoolean(["serial", "autoconnect"], data["autoconnect"])
		settings().save()
		printer.connect(port=port, baudrate=baudrate, profile=printerProfile)
	elif command == "disconnect":
		printer.disconnect()
	elif command == "fake_ack":
		printer.fake_ack()

	return NO_CONTENT

def _get_options():
	connection_options = printer.__class__.get_connection_options()
	profile_options = printerProfileManager.get_all()
	default_profile = printerProfileManager.get_default()

	options = dict(
		ports=connection_options["ports"],
		baudrates=connection_options["baudrates"],
		printerProfiles=[dict(id=printer_profile["id"], name=printer_profile["name"] if "name" in printer_profile else printer_profile["id"]) for printer_profile in profile_options.values() if "id" in printer_profile],
		portPreference=connection_options["portPreference"],
		baudratePreference=connection_options["baudratePreference"],
		printerProfilePreference=default_profile["id"] if "id" in default_profile else None
	)

	return options
